/*
	analog.c trick to use analog inputs as digital inputs
*/
#include "GenericTypeDefs.h"
#include "HardwareProfile.h"
#include "analog.h"
#include "lamps.h"
#include "flash.h"

typedef struct
{
	IoPortId port;
	unsigned int bits;
} port_bit_t;

#if defined(ODYSSEY)
#define	INP_COUNT	0
port_bit_t inputs[INP_COUNT];
#else

#define	INP_COUNT 12

// Inp1, Inp9, Inp10 are analog inputs
port_bit_t inputs[INP_COUNT] = {
	{ IOPORT_A, 0x1 },		// RA0 Inp5
	{ IOPORT_G, 0x80 },		// RG7 Inp6
	{ IOPORT_G, 0x40 },		// RG6 Inp7
	{ IOPORT_C, 0x10 },		// RC4 Inp8
	{ IOPORT_C, 0x8 },		// RC3 Inp9
	{ IOPORT_E, 0x200 },	// RE9 Inp10
	{ IOPORT_C, 0x4 },		// RC2 Inp11
	{ IOPORT_G, 0x8000 },	// RG15 Inp12
	{ IOPORT_C, 0x2 },		// RC1 Inp13
	{ IOPORT_E, 0x80 },		// RE7 Inp14
	{ IOPORT_E, 0x40 },		// RE6 Inp15
	{ IOPORT_E, 0x20 }		// RE5 Inp16
};
#endif

// global variables
INPUT ana_inputs[ANALOG_COUNT];		// state of the analog inputs base on voltage thresholds
INPUT last_inputs[ANALOG_COUNT];	// state of the analog inputs base on voltage thresholds

// local variables 
static volatile unsigned int ADCValue[ANALOG_COUNT];
static volatile BOOL dataReady;
static int tick;
static BOOL patToggle;

void AnalogInit()
{
	int i;

	AD1PCFG = 0xfff0;			// light bar can only use 4 analog inputs
//	AD1PCFG = 0x0;			// AN0-15 are all used, 16 analog inputs
	AD1CON1 = 0x00e0;		//SSRC bit = 111 implies internal
							//counter ends sampling and starts converting.
							//unsigned 10 bit result
//	AD1CHS= 0x00020000;		//Connect RB2/AN2 as CH0 input
							//in this example RB2/AN2 is the input
//	AD1CSSL= 0xffff;		// scan AN0-15, all 16 inputs
	AD1CSSL = 0xf;			// only scan AN0-AN3
	AD1CON3 = 0x0f03;		//Sample time = 15 TAD
	AD1CON2 = 0x040C;		// scan inputs use CSSL. For 4 inputs we must sample every 4th (SMPI = 0x8 = 3 << 2 because 3 = 4th sample)
//	AD1CON2 = 0x043c;		//Select AVDD and AVSS as reference pins
							//Interrupt after every 16th samples
							//scan input selections
	AD1CON1bits.ADON = 1;	//turn ON the ADC
	
	for(i = 0; i < ANALOG_COUNT; i++)
	{
		ana_inputs[i] = OFF;
		last_inputs[i] = OFF;
	}

	// make sure the inputs are set to input
	for(i = 0; i < INP_COUNT; i++)
		PORTSetPinsDigitalIn(inputs[i].port, inputs[i].bits);

	dataReady = FALSE;
	ConfigIntADC10(ADC_INT_ON | ADC_INT_PRI_1);
	tick = GetTickCount();
}

void __ISR(_ADC_VECTOR, ipl1) handlesADCInts(void)
{
	unsigned int na;
	int i, ndx;

	if (IFS1bits.AD1IF)
	{
		AD1CON1bits.ASAM = 0;	//stop sample/convert
		if (! dataReady)		// just in case protection from writing while reading
		{
			// read the ADC values
			ADCValue[2] = ADC1BUF0;		// AN0 is input 3
			ADCValue[1] = ADC1BUF1;		// AN1 is input 2
			ADCValue[0] = ADC1BUF2;		// AN2 is input 1
			ADCValue[3] = ADC1BUF3;		// AN3 is input 4

			// read the digital inputs 2-8, but make them look like analog
#if ! defined(ODYSSEY)
			i = 0;
			for(ndx = 4; ndx < ANALOG_COUNT; ndx++, i++)		// inputs 5-16 are digital, not analog, but simulate
				if (PORTReadBits(inputs[i].port, inputs[i].bits))
					ADCValue[ndx] = 0x150;		// simulate an off value
				else
					ADCValue[ndx] = 0x3ff;		// simulate a high value
#endif
			dataReady = TRUE;
		}
		else
		{
			na = ADC1BUF0;
			na = ADC1BUF1;
			na = ADC1BUF2;
			na = ADC1BUF3;
		}
	}

	IFS1bits.AD1IF = 0;		// clear this int. flag
} // END ADC ISR

static void Start(int ndx, PROG_INPUT *prg)
{
	if (ndx >= CYCLE_INPUTS)
		StartProgram(prg->ProgIndexes[0], prg->Priority, prg->Dimmed, FALSE);
	else
		StartProgram(prg->ProgIndexes[Params.Index[ndx]], prg->Priority, prg->Dimmed, FALSE);
	prg->Toggle = TRUE;
}

static void Stop(int ndx, PROG_INPUT *prg)
{
	if (ndx >= CYCLE_INPUTS)
		StopProgram(prg->ProgIndexes[0]);
	else
		StopProgram(prg->ProgIndexes[Params.Index[ndx]]);
	prg->Toggle = FALSE;
}

void AnalogTasks()
{
	int i, j, highPri, newProg;
	BOOL newDim, done1314;
#if ! defined(ODYSSEY)
	PROG_INPUT *prg;
#endif

	if (Repeat(&tick, 100))		// start an analog sample every 100 ms
		AD1CON1bits.ASAM = 1;	//auto start next sample

	done1314 = FALSE;
	// are the analog samples ready?
	if (dataReady)
	{
		highPri = -1;
#if ! defined(ODYSSEY)
		prg = ProgInputs;
#endif
		for(i = 0; i < ANALOG_COUNT; i++)
		{
			if (ADCValue[i] < 0x70)		// is it low?
				ana_inputs[i] = LOW;
			else
				if (ADCValue[i] > 0x280)		// is it high?
					ana_inputs[i] = HIGH;
				else
					ana_inputs[i] = OFF;			// off or disconnected
// Odyssey has no stand-alone capabilities
#if ! defined(ODYSSEY)
			// for now, the toggle pattern input is 9, since it's analog and works
			if (i == ANALOG_COUNT - 1 && prg != NULL)		// should be ANALOG_COUNT-1 (not 8)
			{
				if (ana_inputs[i] != last_inputs[i])		// change of state?
				{
					if (ana_inputs[i] == HIGH)		// high toggles the pattern
					{
						if (! patToggle)
						{
							// only the 1st four programs can cycle the patterns
							for(j = 0; j < CYCLE_INPUTS; j++)
							{
								if (ProgInputs[j].Toggle)	// is this program running?
								{
									// stop the current pattern
									Stop(j, &ProgInputs[j]);
									// find the next pattern
									while(++Params.Index[j] < MAX_CYCLE)
										if (ProgInputs[j].ProgIndexes[Params.Index[j]] >= 0)	// did we find one?
											break;
									if (Params.Index[j] >= MAX_CYCLE)	// no, reset back to zero
										Params.Index[j] = 0;
									// save this pattern selection
									SaveParams();
									// start the next pattern
									Start(j, &ProgInputs[j]);
								}
							}
							patToggle = TRUE;
						}
					}
					else
						patToggle = FALSE;
				}
			}

			if (prg != NULL && i < MAX_INPUTS)		// only 14 inputs now! 15 is dim input, 16 is cycle,
			{
				// handle timers
				if (prg->InpType == TIMER)		// timer type?
					if (prg->Toggle)
					{
						if (i < CYCLE_INPUTS)
						{
							if (prg->ProgIndexes[Params.Index[i]] >= 0)
								if (prg->RunTimer++ >= prg->TimerValue)
									Stop(i, prg);
						}
						else
						{
							if (prg->ProgIndexes[0] >= 0)
								if (prg->RunTimer++ >= prg->TimerValue)
									Stop(i, prg);
						}
					}
				if (ana_inputs[i] != last_inputs[i])	// input programs and change of state?
				{
					if (prg->ProgIndexes[0] >= 0)		// program for this input
					{
						// added input type logic
						switch(prg->InpType)
						{
							case MOMENTARY:		// momentary input type
								// inputs 13 & 14 can be combined to have a function, normally center out
								if ((i == MAX_INPUTS - 1 || i == MAX_INPUTS - 2) && ! done1314)
								{
									// check if the 13 & 14 input has a program
									if (ProgInputs[MAX_INPUTS].ProgIndexes[0] >= 0)
									{
										// are 13 & 14 high?
										if(ana_inputs[MAX_INPUTS - 1] == HIGH && ana_inputs[MAX_INPUTS - 2] == HIGH)
										{
											// stop program 13
											if (ProgInputs[MAX_INPUTS - 1].ProgIndexes[0] >= 0)
												Stop(MAX_INPUTS - 1, &ProgInputs[MAX_INPUTS - 1]);
											// stop program 14
											if (ProgInputs[MAX_INPUTS - 2].ProgIndexes[0] >= 0)
												Stop(MAX_INPUTS - 2, &ProgInputs[MAX_INPUTS - 2]);
											// Start the input 13 & 14 program
											Start(MAX_INPUTS, &ProgInputs[MAX_INPUTS]);
											done1314 = TRUE;		// we've handled 13 & 14
										}
										else
											Stop(MAX_INPUTS, &ProgInputs[MAX_INPUTS]);	// stop the 13 & 14 program
									}
								}
								if (! done1314)
								{
									if ((prg->Level && ana_inputs[i] == HIGH) || (! prg->Level && ana_inputs[i] == LOW))
										Start(i, prg);
									else
										Stop(i, prg);
								}
								break;
							case TOGGLE:		// toggle on/off
								if ((prg->Level && ana_inputs[i] == HIGH) || (! prg->Level && ana_inputs[i] == LOW))
								{
									if (! prg->IsOn)
									{
										if (! prg->Toggle)
											Start(i, prg);
										else
											Stop(i, prg);
										prg->IsOn = TRUE;
									}
								}
								else
									prg->IsOn = FALSE;
								break;
							case TIMER:		// timer type
								if ((prg->Level && ana_inputs[i] == HIGH) || (! prg->Level && ana_inputs[i] == LOW))
								{
									if (! prg->IsOn)
									{
										prg->RunTimer = 0;
										Start(i, prg);
										prg->IsOn = TRUE;
									}
								}
								else
									prg->IsOn = FALSE;
								break;
						}
					}
				}
				prg++;		// next input
			}
#endif
			last_inputs[i] = ana_inputs[i];		// keep track of last status
		}
		dataReady = FALSE;
	}
}
