/*
	comm.c
*/
#include "GenericTypeDefs.h"
#include "HardwareProfile.h"
#include "timer.h"
#include "lamps.h"
#include "flash.h"
#include "comm.h"
#include "util.h"
#include "analog.h"

#define	BAUDRATE	100000
#define	TX_MAX		32
#define	SOH			1
#define	STX			2
#define	ETX			3
#define	EOT			4
#define	MASTER_ADDRESS	'1'
#define	MY_ADDRESS	'2'

#define	REC_EN		LATDbits.LATD9
#define	DATA_EN		LATDbits.LATD8

static volatile int rxState;
static char rxBuf[RX_MAX+1];
static char txBuf[TX_MAX+1];
static volatile int rxPtr;
static volatile int txLen = 0;
static volatile int txPtr = 0;
static volatile BOOL progFlag;		// new programs to read!
static char expectBlk;
static char lastBlk;
static volatile BOOL powerOnReset = TRUE;

// light bar data fault flags
static BOOL progFault;
volatile BOOL onLine = FALSE;
volatile BOOL watchReset = TRUE;
volatile BOOL forceReset = FALSE;

void CommInit()
{
	rxState = 0;
	progFlag = TRUE;

	TRISDbits.TRISD9 = 0;		// receive enable output
	REC_EN = 0;					// /RE
	TRISDbits.TRISD8 = 0;		// tx enable output
	DATA_EN = 0;				// DE off for now

    // Configure UART module, set buad rate, turn on UART, etc.
    UARTConfigure(UART1, UART_ENABLE_HIGH_SPEED | UART_ENABLE_PINS_TX_RX_ONLY);
	UARTSetFifoMode(UART1, UART_INTERRUPT_ON_TX_NOT_FULL | UART_INTERRUPT_ON_RX_NOT_EMPTY);
    UARTSetLineControl(UART1, UART_DATA_SIZE_8_BITS | UART_PARITY_NONE | UART_STOP_BITS_1);
    UARTSetDataRate(UART1, GetSystemClock(), BAUDRATE);
    UARTEnable(UART1, UART_ENABLE_FLAGS(UART_PERIPHERAL | UART_RX | UART_TX));

    // Configure UART RX Interrupt
    INTEnable(INT_SOURCE_UART_RX(UART1), INT_ENABLED);
	INTEnable(INT_SOURCE_UART_ERROR(UART1), INT_ENABLED);
    INTSetVectorPriority(INT_VECTOR_UART(UART1), INT_PRIORITY_LEVEL_4);
    INTSetVectorSubPriority(INT_VECTOR_UART(UART1), INT_SUB_PRIORITY_LEVEL_0);
}

static void SendCmd(char cmd)
{
	txBuf[0] = SOH;
	txBuf[1] = MASTER_ADDRESS;
	txBuf[2] = cmd;
	txBuf[3] = EOT;
	txPtr = 0;
	txLen = 4;
	DATA_EN = 1;		// enable RS485
    UARTEnable(UART1, UART_ENABLE_FLAGS(UART_PERIPHERAL | UART_TX));
}

static void SendData(char cmd, int len)
{
	txLen = len + 3;
	txBuf[0] = SOH;
	txBuf[1] = MASTER_ADDRESS;
	txBuf[2] = cmd;
	txPtr = 0;
	txBuf[txLen++] = EOT;
	DATA_EN = 1;		// enable RS485
    UARTEnable(UART1, UART_ENABLE_FLAGS(UART_PERIPHERAL | UART_TX));
}

static void ProcessLine()
{
	BOOL progFault = FALSE;
	BOOL ack = TRUE;
	DWORD inp;
	int ndx;

	switch(rxBuf[0])
	{
		case 'K':		// read analog inputs, tristate
			inp = 0;
//			for(ndx = 0; ndx < ANALOG_COUNT; ndx++)
			for(ndx = ANALOG_COUNT - 1; ndx >= 0; ndx--)
			{
				inp <<= 2;
				inp |= ((BYTE)ana_inputs[ndx]) & 0x3;		// 0-Off, 1-Low, 2-High
			}
			ByteToHex(inp >> 24, txBuf + 3);
			ByteToHex(inp >> 16, txBuf + 5);
			ByteToHex(inp >> 8, txBuf + 7);
			ByteToHex(inp & 0xff, txBuf + 9);
			// new, send power on reset
			if (powerOnReset)
			{
				SendCmd('!');		// reset command
				powerOnReset = FALSE;
			}
			else
				SendData('K', 8);
			ack = FALSE;
			break;
		case 'R':		// run a light program
			if (! StartProgram(atoi(rxBuf + 5), atoi(rxBuf + 1), rxBuf[4] == 'T', rxBuf[9] == 'T'))
				progFault = TRUE;
			break;
		case 'S':
			StopProgram(atoi(rxBuf + 1));
			break;
		case 'L':		// start of light bar config
			expectBlk = '0';
			FlashBegin();
		case 'F':
		case 'G':
		case 'P':
		case 'N':
			if (rxBuf[1] != expectBlk)
			{
				if (--expectBlk < '0')
					expectBlk = '9';
				if (expectBlk != rxBuf[1])
					progFault = TRUE;		// block error
			}
			else
				FlashWriteBuf(rxBuf, rxPtr + 1);		// save to flash
			if (++expectBlk > '9')
				expectBlk = '0';
			break;
		case 'Z':
			FlashEnd();
			progFlag = TRUE;	// new programs!
			break;
		case 'W':			// watch dog reset timer
			watchReset = atoi(rxBuf + 1);
			break;
		case 'x':
			forceReset = TRUE;
			break;
		case 'M':
			softDim = rxBuf[1] == 'T';		// software dimmer
			break;
		case 'J':
			runFast = rxBuf[1] == 'T';		// run fast patterns
			break;
	}
	if (progFault)
		SendCmd('P');
	else
		if (ack)
			SendCmd('A');
/*			txBuf[0] = MASTER_ADDRESS;
			txBuf[1] = 'A';
			strcpy(txBuf + 2, fields[fldPtr-1]);
			txPtr = 0;
			txLen = strlen(txBuf);
			txBuf[txLen++] = EOT;
			DATA_EN = 1;		// enable RS485
			UARTSendDataByte(UART1, SOH); */
}

static void ProcessChar(char c)
{
	switch(rxState)
	{
		case 1:
			if (c == MY_ADDRESS)
				rxState++;
			else
				rxState = 0;
			break;
		case 2:
			if (c == ETX)
			{
				rxBuf[rxPtr] = '\0';
				onLine = TRUE;
				ProcessLine();
				rxState = 0;
			}
			else
			{
				if (rxPtr < RX_MAX)
					rxBuf[rxPtr++] = c;
				else
					rxState = 0;		// line too long!
			}
			break;
	}
	if (c == STX)
	{
		rxPtr = 0;
		rxState = 1;
	}
}

void __ISR(_UART_1_VECTOR, ipl4) IntUart1Handler(void)
{
	char c;

	// clear any error interrupts
	if (INTGetFlag(INT_SOURCE_UART_ERROR(UART1)))
	{
		c = U1STA;
		U1STAbits.OERR = 0;
		INTClearFlag(INT_SOURCE_UART_ERROR(UART1));
	}

	if (INTGetFlag(INT_SOURCE_UART_TX(UART1)))
	{
/*		if (txPtr < txLen)
		{
			UARTSendDataByte(UART1, txBuf[txPtr++]);
		}
		else
		{			
			// shut down interrupt
	    	//INTEnable(INT_SOURCE_UART_TX(UART1), INT_DISABLED);
			if (txDone)
			{
				DATA_EN = 0;
				UARTEnable(UART1, UART_ENABLE_FLAGS(UART_PERIPHERAL | UART_RX));	// enable Rx
		    	INTEnable(INT_SOURCE_UART_TX(UART1), INT_DISABLED);
			}
			else
			{
				UARTSetFifoMode(UART1, UART_INTERRUPT_ON_TX_DONE | UART_INTERRUPT_ON_RX_NOT_EMPTY);
				txDone = TRUE;
			}
		} */
		INTClearFlag(INT_SOURCE_UART_TX(UART1));
	}

	// Is this an RX interrupt?
	if(INTGetFlag(INT_SOURCE_UART_RX(UART1)))
	{
		if (UARTReceivedDataIsAvailable(UART1))
		{
			c = UARTGetDataByte(UART1) & 0x7f;
			ProcessChar(c);
		}

		// Clear the RX interrupt Flag
		INTClearFlag(INT_SOURCE_UART_RX(UART1));
	}
}

void CommTasks()
{
	int i;

	if (txLen)
	{
		if (txPtr >= txLen)
		{
			if (UARTTransmissionHasCompleted(UART1))
			{
				DATA_EN = 0;
				txLen = 0;
			    UARTEnable(UART1, UART_ENABLE_FLAGS(UART_PERIPHERAL | UART_RX));
			}
		}
		else
		{
			if(UARTTransmitterIsReady(UART1))
				UARTSendDataByte(UART1, txBuf[txPtr++]);
		}
	}

	if (progFlag)
	{
		ReadPrograms();
		progFlag = FALSE;
	}
}
