/*
	analog.c trick to use analog inputs as digital inputs
*/
#include "GenericTypeDefs.h"
#include "HardwareProfile.h"
#include "analog.h"

// ignition is no longer analog
#define	IGN_INPUT	PORTGbits.RG1

// global variables
INPUT ana_inputs[ANALOG_COUNT];		// state of the analog inputs base on voltage thresholds
INPUT ring_input;					// state of analog horn ring input
unsigned int sysCurrent;			// system current
unsigned int sysVoltage;
int voltInput;						// voltage input to read

// local variables 
static volatile unsigned int ADCValue[ANALOG_COUNT];
static volatile unsigned int anaCurrent;
static volatile unsigned int anaVoltage;
#define		CURR_AVERAGES		16
static int avgs[CURR_AVERAGES];

static volatile BOOL dataReady;
static int tick;

void AnalogInit()
{
	int i;

	AD1PCFG = 0xf0e0;		// AN0-3 and AN8-11 are the 8 inputs, AN4 current sensor
	AD1CON1= 0x00E0;		//SSRC bit = 111 implies internal
							//counter ends sampling and starts converting.
							//unsigned 10 bit result
	AD1CHS= 0x00020000;		//Connect RB2/AN2 as CH0 input
							//in this example RB2/AN2 is the input
	AD1CSSL= 0x0f1f;		// scan AN0-4 and AN8-11 for a total of 9 inputs
	AD1CON3= 0x0f03;		//Sample time = 15 TAD
	AD1CON2= 0x6420;		//Select external VREF+ and VREF- pins
							//Interrupt after every 9th samples
							//scan input selections
	AD1CON1bits.ADON = 1;	//turn ON the ADC
	
	for(i = 0; i < ANALOG_COUNT; i++)
		ana_inputs[i] = OFF;
	ring_input = OFF;
	for(i = 0; i < CURR_AVERAGES; i++)
		avgs[i] = 0;
	dataReady = FALSE;
	voltInput = 0;
	sysCurrent = 512;
	sysVoltage = 0;
	ConfigIntADC10(ADC_INT_ON | ADC_INT_PRI_1);
	tick = GetTickCount();
}

void __ISR(_ADC_VECTOR, ipl1) handlesADCInts(void)
{
	unsigned int tmp;

	if (IFS1bits.AD1IF)
	{
		AD1CON1bits.ASAM = 0;	//stop sample/convert
		if (! dataReady)		// just in case protection from writing while reading
		{
			// read the ADC values
			ADCValue[4] = ADC1BUF0;
			ADCValue[5] = ADC1BUF1;
			ADCValue[6] = ADC1BUF2;
			ADCValue[7] = ADC1BUF3;
			anaCurrent = ADC1BUF4;
			ADCValue[3] = ADC1BUF5;
			ADCValue[2] = ADC1BUF6;
			ADCValue[1] = ADC1BUF7;
			ADCValue[0] = ADC1BUF8;			// this was ignition, but is now horn ring
			// now hard coded to input 3 as per D & R
			anaVoltage = ADCValue[2];
//			anaVoltage = ADCValue[voltInput];
			dataReady = TRUE;
		}
		else
		{
			tmp = ADC1BUF0;
			tmp = ADC1BUF1;
			tmp = ADC1BUF2;
			tmp = ADC1BUF3;
			tmp = ADC1BUF4;
			tmp = ADC1BUF5;
			tmp = ADC1BUF6;
			tmp = ADC1BUF7;
			tmp = ADC1BUF8;
		}
	}

	IFS1bits.AD1IF = 0;		// clear this int. flag
} // END ADC ISR

/*
	average 8 readings of current as per D & R
*/
static int Average(int n)
{
	int i, tot;

	for(i = 0; i < CURR_AVERAGES - 1; i++)
		avgs[i] = avgs[i+1];
	avgs[CURR_AVERAGES-1] = n;	
	tot = 0;
	for(i = 0; i < CURR_AVERAGES; i++)
		tot += avgs[i];
	return tot / CURR_AVERAGES;			
}

void AnalogTasks()
{
	int i;

	if (Repeat(&tick, 100))		// start an analog sample every 100 ms
		AD1CON1bits.ASAM = 1;	//auto start next sample

	// are the analog samples ready?
	if (dataReady)
	{
		// ignition input is no longer analog
		if (IGN_INPUT)
			ana_inputs[0] = HIGH;
		else
			ana_inputs[0] = OFF;
		// horn ring is analog 11 (ADCValue[0]]
		if (ADCValue[0] < 0x70)		// is it low?
			ring_input = LOW;
		else
			if (ADCValue[0] > 0x300)		// is it high?
				ring_input = HIGH;
			else
				ring_input = OFF;			// off or disconnected
		
		// start at 1 since zero is horn ring
		for(i = 1; i < ANALOG_COUNT; i++)
		{
			if (ADCValue[i] < 0x70)		// is it low?
				ana_inputs[i] = LOW;
			else
				if (ADCValue[i] > 0x300)		// is it high?
					ana_inputs[i] = HIGH;
				else
					ana_inputs[i] = OFF;			// off or disconnected
		}

		sysCurrent = Average(anaCurrent);
		//sysCurrent = anaCurrent;
		sysVoltage = anaVoltage;
		dataReady = FALSE;
	}
}
