/*
	comm.c
*/
#include "GenericTypeDefs.h"
#include "HardwareProfile.h"
#include "timer.h"
#include "comm.h"
#include "sound.h"
#include "analog.h"
#include "util.h"
#include "flash.h"
#include "lamps.h"

#define	BAUDRATE	100000
#define	TX_MAX		128
#define	SOH			1
#define	STX			2
#define	ETX			3
#define	EOT			4
#define	MASTER_ADDRESS	'1'
#define	MY_ADDRESS	'3'

#define	REC_EN		LATDbits.LATD9
#define	DATA_EN		LATDbits.LATD8

static volatile int rxState;
static char rxBuf[RX_MAX+1];
static char txBuf[TX_MAX+1];
static volatile int rxPtr;
static volatile int txLen = 0;
static volatile int txPtr = 0;
static volatile char expectBlk;
static volatile BOOL progFlag = TRUE;
static int hornTick;
static int hornState = 0;
static int hornCount;
// horn status that is sent
static volatile BOOL hornPulse = FALSE;
static BOOL hornOn = FALSE;
static BOOL powerUp = TRUE;		// powered up flag

// global for Comm failure
volatile BOOL onLine = FALSE;
volatile BOOL watchReset = TRUE;
volatile BOOL forceReset = FALSE;
BOOL ActiveLowHorn = FALSE;

#define		INP_COUNT		9

port_bit_t outputs[OUT_COUNT] = {
	{ IOPORT_A, BIT_7 },		// RA7 Out1
	{ IOPORT_A, BIT_6 },		// RA6 Out2
	{ IOPORT_E, BIT_0 },		// RE0 Out3
	{ IOPORT_E, BIT_1 },		// RE1 Out4
	{ IOPORT_G, BIT_14 },	// RG14 Out5
	{ IOPORT_G, BIT_12 },	// RG12 Out6
	{ IOPORT_G, BIT_13 },	// RG13 Out7
	{ IOPORT_E, BIT_2 },		// RE2 Out8
	{ IOPORT_G, BIT_6 },		// RG6 Out9
	{ IOPORT_G, BIT_7 },		// RG7 Out10
	{ IOPORT_C, BIT_4 },		// RC4 Out11
	{ IOPORT_C, BIT_3 },		// RC3 Out12
	{ IOPORT_C, BIT_1 },		// RC1 Out13
	{ IOPORT_C, BIT_2 },		// RC2 Out14
	{ IOPORT_E, BIT_7 },		// RE7 Out15
	{ IOPORT_E, BIT_6 },		// RE6 Out16
	{ IOPORT_E, BIT_5 },		// RE5 Out20
	{ IOPORT_G, BIT_15 },	// RG15 Out19
	{ IOPORT_E, BIT_4 },		// RE4 Out18
	{ IOPORT_E, BIT_3 },		// RE3 Out17
// expansion 1
	{ IOPORT_G, 0x100 },	// RG8 Out21
	{ IOPORT_G, 0x200 },	// RG9 Out22
	{ IOPORT_A, 0x1 },		// RA0 Out23
	{ IOPORT_E, 0x100 },		// RE8 Out24
	{ IOPORT_A, 0x2 },		// RA1 Out25
	{ IOPORT_F, 0x2000 },	// RF13 Out26
	{ IOPORT_F, 0x1000 },	// RF12 Out27
	{ IOPORT_B, 0x1000 },	// RB12 Out28
	{ IOPORT_B, 0x2000 },	// RB13 Out29
	{ IOPORT_B, 0x4000 },	// RB14 Out30
// expansion 2
	{ IOPORT_D, 0x8000 },	// RD15 Out31
	{ IOPORT_F, 0x10 },		// RF4 Out32
	{ IOPORT_F, 0x20 },		// RF5 Out33
	{ IOPORT_A, 0x4 },		// RA2 Out34
	{ IOPORT_A, 0x8 },		// RA3 Out35
	{ IOPORT_A, 0x10 },		// RA4 Out36
	{ IOPORT_A, 0x20 },		// RA5 Out37
	{ IOPORT_D, 0x4 },		// RD2 Out38
	{ IOPORT_D, 0x8 },		// RD3 Out39
	{ IOPORT_D, 0x1000 }	// RD12 Out40
};

void CommInit()
{
	int ndx;

	rxState = 0;
	powerUp = TRUE;

	// use RD10 pin 70 as the power output
	TRISDbits.TRISD10 = 0;
	LATDbits.LATD10 = 1;

	TRISDbits.TRISD9 = 0;		// receive enable output
	REC_EN = 0;					// /RE
	TRISDbits.TRISD8 = 0;		// tx enable output
	DATA_EN = 0;				// DE off for now

	for(ndx = 0; ndx < OUT_COUNT; ndx++)
	{
		PORTSetPinsDigitalOut(outputs[ndx].port, outputs[ndx].bits);
		OutOff(ndx);		// PORTClearBits(outputs[ndx].port, outputs[ndx].bits);
	}

    // Configure UART module, set buad rate, turn on UART, etc.
    UARTConfigure(UART1, UART_ENABLE_HIGH_SPEED | UART_ENABLE_PINS_TX_RX_ONLY);
	UARTSetFifoMode(UART1, UART_INTERRUPT_ON_RX_NOT_EMPTY);
    UARTSetLineControl(UART1, UART_DATA_SIZE_8_BITS | UART_PARITY_NONE | UART_STOP_BITS_1);
    UARTSetDataRate(UART1, GetSystemClock(), BAUDRATE);
    UARTEnable(UART1, UART_ENABLE_FLAGS(UART_PERIPHERAL | UART_RX));

    // Configure UART RX Interrupt
    INTEnable(INT_SOURCE_UART_RX(UART1), INT_ENABLED);
	INTEnable(INT_SOURCE_UART_ERROR(UART1), INT_ENABLED);
	INTSetVectorPriority(INT_VECTOR_UART(UART1), INT_PRIORITY_LEVEL_4);
	INTSetVectorSubPriority(INT_VECTOR_UART(UART1), INT_SUB_PRIORITY_LEVEL_0);

	hornTick = GetTickCount();
}

static void SendCmd(char cmd)
{
	txBuf[0] = SOH;	// send 2, one is a padding character
	txBuf[1] = SOH;
	txBuf[2] = MASTER_ADDRESS;
	txBuf[3] = cmd;
	txBuf[4] = EOT;
	txPtr = 0;
	txLen = 5;
	DATA_EN = 1;		// enable RS485
    UARTEnable(UART1, UART_DISABLE_FLAGS(UART_PERIPHERAL | UART_RX));
    UARTEnable(UART1, UART_ENABLE_FLAGS(UART_PERIPHERAL | UART_TX));
}

static void SendData(char cmd, int len)
{
	txLen = len + 4;
	txBuf[0] = SOH;
	txBuf[1] = SOH;
	txBuf[2] = MASTER_ADDRESS;
	txBuf[3] = cmd;
	txPtr = 0;
	txBuf[txLen++] = EOT;
	DATA_EN = 1;		// enable RS485
    UARTEnable(UART1, UART_DISABLE_FLAGS(UART_PERIPHERAL | UART_RX));
    UARTEnable(UART1, UART_ENABLE_FLAGS(UART_PERIPHERAL | UART_TX));
}

void OutOn(int ndx)
{
	if (ndx >= 0 && ndx < OUT_COUNT)
	{
/*		if (ndx > 15)
		{
			switch(ndx)
			{
				case 16:		// out 17
					LATEbits.LATE5 = 1;
					break;
				case 17:		// out 18
					LATGbits.LATG15 = 1;
					break;
				case 18:		// out 19
					LATEbits.LATE4 = 1;
					break;
				case 19:		// out 20
					LATEbits.LATE3 = 1;
					break;
			}
		}
		else */
			PORTSetBits(outputs[ndx].port, outputs[ndx].bits);
	}
}

void OutOff(int ndx)
{
	if (ndx >= 0 && ndx < OUT_COUNT)
		PORTClearBits(outputs[ndx].port, outputs[ndx].bits);
}

static void ProcessCmd()
{
	int ndx, msk;
	DWORD inp;
	int n;
	BOOL ack = TRUE;
	BOOL b;
	BOOL progFault = FALSE;

	switch(*rxBuf)
	{
		case 'O':		// output on
			OutOn(atoi(rxBuf + 1));
			break;
		case 'F':		// output off
			OutOff(atoi(rxBuf + 1));
			break;
		case 'X':		// all off
			for(ndx = 0; ndx < OUT_COUNT; ndx++)
				PORTClearBits(outputs[ndx].port, outputs[ndx].bits);
			break;
/*		case 'I':		// read inputs
			inp = 0;
			msk = 1;
			for(ndx = 0; ndx < INP_COUNT; ndx++, msk <<= 1)
				if (PORTReadBits(inputs[ndx].port, inputs[ndx].bits))
					inp |= msk;
			ByteToHex(inp >> 8, txBuf + 3);
			ByteToHex(inp & 0xff, txBuf + 5);
			SendData('I', 4);
			ack = FALSE;
			break; */
		case 'V':		// read the analog system current
			// set the voltage input
			ndx = atoi(rxBuf + 1);
			if (ndx >= 8)	// protection
				ndx = 0;
			voltInput = ndx;
			itoan(sysCurrent, txBuf + 4, 4);
			txBuf[8] = ',';
			itoan(sysVoltage, txBuf + 9, 4);
			SendData('V', 9);
			ack = FALSE;
			break;
		case 'P':		// ask if just powered on
			if (powerUp)
			{
				powerUp = FALSE;
				SendData('Y', 0);
			}
			else
				SendData('A', 0);
			ack = FALSE;
			break;
		case 'K':		// read analog tristate inputs
			inp = 0;
			for(ndx = ANALOG_COUNT; ndx >= 0; ndx--)
			{
				inp <<= 2;
				inp |= ((BYTE)ana_inputs[ndx]) & 0x3;		// 0-Off, 1-Low, 2-High
			}
			// handle horn ring specially
/* we don't send this anymore!
			if (hornOn)
				inp |= 2 << (8 * 2);			// set input 8 to high
*/
			// pulse is only sent once
			if (hornPulse)
			{
				inp |= 2 << (8 * 2);		// pulse is now 8 instead of 9
				hornPulse = FALSE;		// only once
			}	
			ByteToHex(inp >> 24, txBuf + 4);
			ByteToHex(inp >> 16, txBuf + 6);
			ByteToHex(inp >> 8, txBuf + 8);
			ByteToHex(inp & 0xff, txBuf + 10);
			SendData('K', 8);
			ack = FALSE;
			break;
		case 'B':		
			// sound bit, 1st byte is siren 1 or 2, 2nd byte true is on, 3nd byte is bit
			ndx = rxBuf[1] & 1;
			b = rxBuf[2] != '0';
			n = atoi(rxBuf + 3);
			switch(n)
			{
				case 1:
					StandByMode = FALSE;
					AirHornBit[ndx] = b;
					if (ndx == 0)
						AirHornBit[1] = b;		// as per D&R, turn it on to the secondary amp as well
					break;
				case 2:
					StandByMode = FALSE;
					if (! b)
					{
						ManualBit[ndx] = FALSE;
						ManualStop[ndx] = TRUE;
					}
					else
					{
						ManualStart[ndx] = TRUE;
						ManualStop[ndx] = FALSE;
						ManualBit[ndx] = TRUE;
					}
					break;
				case 3:
				case 4:
					CancelStandbyTone(0, TRUE);		// kill any previous standby stuff
					StandByMode = b;		// new, standby mode
					StandByTone = n;
					if (! b)		// is tone going off?
					{
						if (StandByTone == 3)	// air horn must go off
						{
							AirHornBit[0] = FALSE;
							AirHornBit[1] = FALSE;
						}
						else
						{
							ManualBit[0] = FALSE;
							ManualStop[0] = FALSE;
							ManualBit[1] = FALSE;
							ManualStop[1] = FALSE;
						}
					}
					break;
			}
			break;
		case 'M':		// siren
			// 1st byte is siren 1 or 2, second is sound
			ndx = rxBuf[1] & 1;
			SirenControl[ndx] = atoi(rxBuf + 2);
			StandByMode = FALSE;		// must be out of standby
			break;
		case 'T':		// tone mode
			// 1st byte is siren 1 or 2, second is tone
			ndx = rxBuf[1] & 1;
			toneMode[ndx] = atoi(rxBuf + 2);
			break;
		case 'R':		// run a light program
			if (! StartProgram(atoi(rxBuf + 5), atoi(rxBuf + 1), rxBuf[4] == 'T'))
				progFault = TRUE;
			break;
		case 'S':
			StopProgram(atoi(rxBuf + 1));
			break;
		case 't':		// start of traffic director config
			expectBlk = '0';
			FlashBegin();
		case 'f':
		case 'g':
		case 'p':
		case 'n':
			if (rxBuf[1] != expectBlk)
			{
				if (--expectBlk < '0')
					expectBlk = '9';
				if (expectBlk != rxBuf[1])
					progFault = TRUE;		// block error
			}
			else
				FlashWriteBuf(rxBuf, rxPtr + 1);		// save to flash
			if (++expectBlk > '9')
				expectBlk = '0';
			break;
		case 'Z':
			FlashEnd();
			progFlag = TRUE;	// new programs!
			break;
		case 'Q':		// power down the system!
			LATDbits.LATD10 = 0;		// power down!
			break;
		case 'W':			// watch dog reset timer
			watchReset = atoi(rxBuf + 1);
			break;
		case 'x':
			forceReset = TRUE;
			break;
	}
	if (progFault)
		SendCmd('P');
	else
		if (ack)
			SendCmd('A');
}

static void ProcessChar(char c)
{
	switch(rxState)
	{
		case 1:
			if (c == MY_ADDRESS)
				rxState++;
			else
				rxState = 0;
			break;
		case 2:
			if (c == ETX)
			{
				rxBuf[rxPtr] = '\0';
				onLine = TRUE;
				ProcessCmd();
				rxState = 0;
			}
			else
			{
				if (rxPtr < RX_MAX)
					rxBuf[rxPtr++] = c;
				else
					rxState = 0;		// line too long!
			}
			break;
	}
	if (c == STX)
	{
		rxPtr = 0;
		rxState = 1;
	}
}

void __ISR(_UART_1_VECTOR, ipl4) IntUart1Handler(void)
{
	char c;

	if (INTGetFlag(INT_SOURCE_UART_ERROR(UART1)))
	{
		c = U1STA;
		U1STAbits.OERR = 0;
		INTClearFlag(INT_SOURCE_UART_ERROR(UART1));
	}

	if (INTGetFlag(INT_SOURCE_UART_TX(UART1)) )
	{
		INTClearFlag(INT_SOURCE_UART_TX(UART1));
	}

	// Is this an RX interrupt?
	if(INTGetFlag(INT_SOURCE_UART_RX(UART1)))
	{
//		if (INTGetEnable(INT_SOURCE_UART_RX(UART1)))
		{
			while (UARTReceivedDataIsAvailable(UART1))
			{
				c = UARTGetDataByte(UART1) & 0x7f;
				ProcessChar(c);
			}
		}
		// Clear the RX interrupt Flag
		INTClearFlag(INT_SOURCE_UART_RX(UART1));
	}

}

/*
	handle active low or high horn ring input
*/
BOOL HornRing()
{
	return ActiveLowHorn ? ring_input == LOW : ring_input == HIGH;
}

void CommTasks()
{
	int i;

	if (txLen)
	{
		if (txPtr >= txLen)
		{
			if (UARTTransmissionHasCompleted(UART1))
			{
				DATA_EN = 0;
				txLen = 0;
			    UARTEnable(UART1, UART_DISABLE_FLAGS(UART_PERIPHERAL | UART_TX));
			    UARTEnable(UART1, UART_ENABLE_FLAGS(UART_PERIPHERAL | UART_RX));
			}
		}
		else
		{
			if(UARTTransmitterIsReady(UART1))
				UARTSendDataByte(UART1, txBuf[txPtr++]);
		}
	}
	if (progFlag)
	{
		progFlag = FALSE;
		ReadPrograms();
	}

	// horn ring logic
	if (Repeat(&hornTick, 50))		// every 50ms
	{
		if (! StandByMode)		// not in standby!
		{
			switch(hornState)
			{
				case 0:
					if (HornRing())		// wait for horn ring on
					{
						hornPulse = FALSE;	// it may have been left on
						hornState++;
						hornCount = 0;
					}
					break;
				case 1:
					if (! HornRing())		// if the horn ring goes off, then indicate a pulse
					{
						hornPulse = TRUE;
						hornState = 0;
					}
					else
					{
						if (++hornCount >= 10)		// 500 ms		
						{
							if (SirenControl[0] != NO_DEVICE)
							{
								AirHornBit[0] = TRUE;	// turn on the air horn as per D&R
								if (SirenControl[1] != NO_DEVICE)
									AirHornBit[1] = TRUE;	// turn on both as per D&R
							}
							hornState++;
						}
					}
					break;
				case 2:
					if (! HornRing())		// if the horn ring goes off, then turn off hornOn
					{
						if (SirenControl[0] != NO_DEVICE)
						{
							AirHornBit[0] = FALSE;	// turn off the air horn as per D&R
							if (SirenControl[1] != NO_DEVICE)
								AirHornBit[1] = FALSE;	// turn off the air horn as per D&R
						}
						hornState = 0;
					}
					break;
			}
		}
	}
}
