/*
	lamps.c
	D&R Electronics Lamp Controller Lamp outputs
*/
#include "GenericTypeDefs.h"
#include "HardwareProfile.h"
#include "comm.h"
#include "lamps.h"

/* locals */

static int currProgNdx;
static BOOL progDimmed;
static int dimDelay = 0;

static BOOL lampIsDim[OUT_COUNT];

PROG_LIST Programs = { 0 };	// structure of programs, default to NO programs!
PROG_INPUT *ProgInputs = NULL;		// program inputs for stand-alone bar

static void EnableTimer5()
{
	// This statement says: turn on timer3 | have it use an internal clock source | have it
	// use a prescaler of 1:256, and use a period of 234, give 500 ints / sec for 2ms
	OpenTimer5( T5_ON | T5_SOURCE_INT | T5_PS_1_256, 233);// 585);		// approx 2 ms interrupt
	
	// Timer2 as configured would trigger an interrupt at a frequency of (80MHZ/256/65535), or 4.77
	// times a second.
	
	// This statement configured the timer to produce an interrupt with a priority of 3
	ConfigIntTimer5( T5_INT_ON | T5_INT_PRIOR_3);
}

static void LampOn(int lamp)
{
	if (lamp < OUT_COUNT)
		PORTSetBits(outputs[lamp].port, outputs[lamp].bits);
}

static void LampOff(int lamp)
{
	if (lamp < OUT_COUNT)
		PORTClearBits(outputs[lamp].port, outputs[lamp].bits);
}

static void SetLamp(int lamp, BOOL on)
{
	if (lamp < OUT_COUNT)
	{
		if (on)
			PORTSetBits(outputs[lamp].port, outputs[lamp].bits);
		else
			PORTClearBits(outputs[lamp].port, outputs[lamp].bits);
	}
}

void LampInit(void)
{
	int ndx, lamp;
	PROGRAM *cProg;

	// stop all programs
	for(ndx = 0; ndx < Programs.count; ndx++)
	{
		cProg = &Programs.progs[ndx];
		cProg->Running = FALSE;
	}
	for(lamp = 0; lamp < OUT_COUNT; lamp++)
	{
		PORTSetPinsDigitalOut(outputs[lamp].port, outputs[lamp].bits);
		LampOff(lamp);
		lampIsDim[lamp] = FALSE;	// dimming logic bit
	}
	EnableTimer5();
}

/****************************************************************************
  Function:
    void __attribute__((__interrupt__, auto_psv)) _T2Interrupt(void)

  Description:
    Timer ISR, used for PWM for brightness
  Precondition:
    None

  Return Values:
    None

  ***************************************************************************/

void __ISR(_TIMER_5_VECTOR, ipl3) handlesTimer5Ints(void)
{
	int i, prNdx, pNdx, lNdx;
	PROGRAM *cProg;
	LAMP_PROGRAM *lProg;
	LAMP_GROUP *lGroup;
	LAMP_SEQ *lSeq;
	int lmp;
	BOOL running = FALSE;
//	LAMP *lmp;

	for(prNdx = 0; prNdx < Programs.count; prNdx++)
	{
		cProg = &Programs.progs[prNdx];
		if (cProg->Running)
		{
			running = TRUE;
			for(pNdx = 0; pNdx < cProg->count; pNdx++)
			{
				lProg = &cProg->lamp_programs[pNdx];
				lProg->CurrentDelay += 2;		//  2 ms time base
				if (lProg->CurrentDelay > lProg->NextDelay)
				{
					lGroup = &lProg->lamp_groups[lProg->PatternNdx];
	//	            PatternSimple ps = pc.Patterns[pc.PatternNdx];
		            if (lProg->LampNdx >= lGroup->count)	// pc.LampNdx >= ps.Lamps.Length)
		            {
		                if (++lProg->RepeatCnt >= lGroup->rep)
		                {
		                    lProg->RepeatCnt = 0;
		                    if (++lProg->PatternNdx >= lProg->count)
		                        lProg->PatternNdx = 0;
							lGroup = &lProg->lamp_groups[lProg->PatternNdx];
		                }
		                lProg->LampNdx = 0;
		            }
					lSeq = &lGroup->lamp_seqs[lProg->LampNdx++];
		            //LampSimple ls = ps.Lamps[pc.LampNdx++];
		
					for(lNdx = 0; lNdx < lSeq->count; lNdx++)
					{
						lmp = lSeq->lamps[lNdx].lamp;
						if(cProg->LampEn[lmp])		// lower priority check
						{
							// if this program isn't dimmed, or if this lamp cannot dim, turn it on
							if (! cProg->Dimmed || ! Programs.dimmable[lmp])
								SetLamp(lmp, lSeq->lamps[lNdx].on);
							else
							{
								lampIsDim[lmp] = lSeq->lamps[lNdx].on;		// lamp is dim, just set a flag
								LampOff(lmp);		// and turn it off
							}
						}
					}
					lProg->NextDelay = lSeq->delay;	
					lProg->CurrentDelay = 0;
				}
			}
		}
	}
	// only if something is running
	if (running)
	{
		// dimming logic
		// use a pwm for this!
		if (dimDelay > 3)
		{
			for(i = 0; i < OUT_COUNT; i++)
				if (lampIsDim[i])
					LampOn(i);
			dimDelay = 0;
		}
		else
		{
			if (dimDelay++ == 0)
				for(i = 0; i < OUT_COUNT; i++)
					if (lampIsDim[i])
						LampOff(i);
		}
	}

	mT5ClearIntFlag();

} // END Timer5 ISR


static void CalcPriority()
{
	int pri, lmp, pgm;
	PROGRAM *cProg;

	CloseTimer5();		// stop logic from running
	// check all programs that are running
	for(lmp = 0; lmp < OUT_COUNT; lmp++)
	{
		pri = -1;
		// find the highest priority number for this lamp
		for(pgm = 0; pgm < Programs.count; pgm++)
		{
			cProg = &Programs.progs[pgm];
			cProg->LampEn[lmp] = FALSE;		// disable all lamps
			if (cProg->Running && cProg->LampList[lmp])
			{
				if (cProg->Priority > pri)
					pri = cProg->Priority;
			}
		}
		// is this lamp on anywhere?
		if (pri != -1)
		{
			// loop through again to set the lamp with the highest priority
			for(pgm = 0; pgm < Programs.count; pgm++)
			{
				cProg = &Programs.progs[pgm];
				// set the highest priority lamp we can find
				if (cProg->Running && cProg->LampList[lmp] && cProg->Priority == pri)
				{
					cProg->LampEn[lmp] = TRUE;
					break;
				}
			}
		}
	}
	EnableTimer5();		// restart logic
}

void StopProgram(int pgm)
{
	int i;
	PROGRAM *cProg;

	if (pgm < Programs.count)
	{
		cProg = &Programs.progs[pgm];
		if (cProg->Running)		// ignore if not running
		{
			cProg->Running = FALSE;
			for(i = 0; i < OUT_COUNT; i++)
			{
				if(cProg->LampEn[i])		// if this lamp enabled to run
				{
					lampIsDim[i] = FALSE;	// dimming logic bit
					LampOff(i);
				}
			}
			// recalculate lamp priority
			CalcPriority();
		}
	}
}

BOOL StartProgram(int pgm, int pri, BOOL dim)
{
	int i;
	BOOL rc = FALSE;
	PROGRAM *cProg;

	if (pgm < Programs.count)		//is it valid?
	{
		rc = TRUE;
		cProg = &Programs.progs[pgm];		// pointer to current program
		if (! cProg->Running)		// ignore if already running
		{
			cProg->Dimmed = dim;
			cProg->Priority = pri;
			for(i = 0; i < cProg->count; i++)
			{
				cProg->lamp_programs[i].PatternNdx = 0;
				cProg->lamp_programs[i].LampNdx = 0;
				cProg->lamp_programs[i].RepeatCnt = 0;
				cProg->lamp_programs[i].NextDelay = 0;
				cProg->lamp_programs[i].CurrentDelay = 0;
			}
			cProg->Running = TRUE;
			CalcPriority();
		}
	}	
	return rc;
}

