/*
	comm.c
	comment
*/
#define _SUPPRESS_PLIB_WARNING 1 

#include "GenericTypeDefs.h"
#include "HardwareProfile.h"
#include "timer.h"
#include "comm.h"
#include "logic.h"

#define	BAUDRATE	100000
#define	TX_MAX		1024
#define	RX_MAX		32
#define	MAX_FIELDS	10

#define	SOH			1
#define	STX			2
#define	ETX			3
#define	EOT			4

#define	REC_EN		LATDbits.LATD9
#define	DATA_EN		LATDbits.LATD8

/* locals */
static char *fields[MAX_FIELDS+1];
static volatile int fldPtr;
static volatile int rxState;
static volatile int rxPtr;
static volatile char rxResponse;
static volatile BOOL rxDone;
static volatile int txPtr;
static volatile int txLen;
static char txBuf[TX_MAX];

/* globals */
char rxBuf[RX_MAX+1];

void CommInit()
{
	rxState = 0;

	TRISDbits.TRISD9 = 0;		// receive enable output
	REC_EN = 0;					// /RE
	TRISDbits.TRISD8 = 0;		// tx enable output
	DATA_EN = 0;				// DE off for now

    // Configure UART module, set buad rate, turn on UART, etc.
    UARTConfigure(UART1, UART_ENABLE_HIGH_SPEED | UART_ENABLE_PINS_TX_RX_ONLY);
	UARTSetFifoMode(UART1, UART_INTERRUPT_ON_RX_NOT_EMPTY);
    UARTSetLineControl(UART1, UART_DATA_SIZE_8_BITS | UART_PARITY_NONE | UART_STOP_BITS_1);
    UARTSetDataRate(UART1, GetSystemClock(), BAUDRATE);
    UARTEnable(UART1, UART_ENABLE_FLAGS(UART_PERIPHERAL | UART_RX));

    // Configure UART RX Interrupt
    INTEnable(INT_SOURCE_UART_RX(UART1), INT_ENABLED);
	INTEnable(INT_SOURCE_UART_ERROR(UART1), INT_ENABLED);
	//IEC0bits.U1EIE = 1;		// enable error interrupt
    INTSetVectorPriority(INT_VECTOR_UART(UART1), INT_PRIORITY_LEVEL_4);
    INTSetVectorSubPriority(INT_VECTOR_UART(UART1), INT_SUB_PRIORITY_LEVEL_0);
}

static void StartSend(char addr, char *buf, int len)
{
	int i, st;

	DATA_EN = 1;		// enable RS485
	txBuf[0] = STX;
	txBuf[1] = STX;
	txBuf[2] = addr;
	memmove(txBuf + 3, buf, len);
	txBuf[len + 3] = ETX;
	rxState = 0;
	rxDone = FALSE;
	txPtr = 0;
	txLen = len + 4;		// must include the stx, addr & etx
    UARTEnable(UART1, UART_DISABLE_FLAGS(UART_PERIPHERAL | UART_RX));
    UARTEnable(UART1, UART_ENABLE_FLAGS(UART_PERIPHERAL | UART_TX));
}

static void SendData()
{
	if (txLen)
	{
		if (txPtr >= txLen)
		{
			if (UARTTransmissionHasCompleted(UART1))
			{
				DATA_EN = 0;
				txLen = 0;
			    UARTEnable(UART1, UART_DISABLE_FLAGS(UART_PERIPHERAL | UART_TX));
			    UARTEnable(UART1, UART_ENABLE_FLAGS(UART_PERIPHERAL | UART_RX));
			}
		}
		else
		{
			if(UARTTransmitterIsReady(UART1))
				UARTSendDataByte(UART1, txBuf[txPtr++]);
		}
	}
}

char SendLine(char addr, char *buf, int len)
{
	return SendLineEx(addr, buf, len, 100);
}

char SendLineEx(char addr, char *buf, int len, int delay)
{
	int retry, tick;

	retry = 3;
	rxResponse = 'N';		// force to NAK, no response
	while(retry--)
	{
		ClearWDT();
		StartSend(addr, buf, len);
		tick = GetTickCount();
		while(! rxDone)
		{
			ClearWDT();
			if (HasElapsed(tick, delay))
				break;
			SendData();
		}
		if (rxDone)
			break;
	}
	return rxResponse;
}

void ForceReset(BYTE addr)
{
	int tick;
	char s[3];

	s[0] = 'x';
	StartSend(addr, s, 1);
	tick = GetTickCount();
	while(! HasElapsed(tick, 250))
		ClearWDT();
}

BOOL WatchReset(BYTE addr, BOOL on)
{
	char s[8];

	s[0] = 'W';
	if (on)
		s[1] = '1';
	else
		s[1] = '0';
	s[2] = '\0';
	return SendLine(addr, s, 2) == 'A';
}

BOOL RunLightProgram(BYTE addr, int pgm, int pri, BOOL dim, BOOL canfast)
{
	char s[16];

	s[0] = 'R';
	itoan(pri, s + 1, 3);
	s[4] = dim ? 'T' : 'F';
	itoan(pgm, s + 5, 4);
	s[9] = canfast ? 'T' : 'F';
	s[10] = '\0';
	return SendLine(addr, s, 10) == 'A';
}

BOOL StopLightProgram(BYTE addr, int pgm)
{
	char s[8];

	s[0] = 'S';
	itoan(pgm, s + 1, 4);
	return SendLine(addr, s, 5) == 'A';
}

BOOL FastPatterns(BYTE addr, BOOL on)
{
	char s[3];

	s[0] = 'J';
	s[1] = on ? 'T' : 'F';
	return SendLine(addr, s, 2) == 'A';
}

BOOL SoftwareDimmer(BYTE addr, BOOL on)
{
	char s[3];

	s[0] = 'M';
	s[1] = on ? 'T' : 'F';
	return SendLine(addr, s, 2) == 'A';
}

BOOL ReadInputs(BYTE addr, BYTE cmd)
{
	char s[8];

	s[0] = cmd;
	s[1] = '\0';
	return SendLine(addr, s, 1) == cmd;
}

BOOL ReadVoltage(BYTE addr)
{
	char s[8];

	s[0] = 'V';
	s[1] = TouchPad.BattVoltInput + '0';
	s[2] = '\0';
	return SendLine(addr, s, 2) == 'V';
}

BOOL ReadTemperature(BYTE addr)
{
	char s[6];

	s[0] = 'T';
	s[1] = '\0';
	return SendLine(addr, s, 1) == 'T';
}

BOOL SendPowerOff()
{
	char s[2];

	s[0] = 'Q';
	return SendLine(RELAY_ADDRESS, s, 1) == 'A';
}


BOOL Relay(int rly, BOOL on)
{
	char s[8];
	BOOL rc = TRUE;

	if (on)
		s[0] = 'O';
	else
		s[0] = 'F';
	itoan(rly, s + 1, 2);
	if (TouchPad.RBExists)
		rc = SendLine(RELAY_ADDRESS, s, 3) == 'A';
	return rc;
}

BOOL SoundControl(int ndx, int siren)
{
	char s[8];
	BOOL rc = TRUE;

	s[0] = 'M';
	s[1] = '0' + ndx;
	itoan(siren, s + 2, 2);
	if (TouchPad.RBExists)
		rc = SendLine(RELAY_ADDRESS, s, 4) == 'A';
	return rc;
}

BOOL RelayBoxPowerUp()
{
	BOOL rc = FALSE;
	char s[6];

	s[0] = 'P';
	s[1] = '\0';
	if (TouchPad.RBExists)
		rc = SendLine(RELAY_ADDRESS, s, 1) == 'Y';
	return rc;
}

BOOL SoundBit(int ndx, int b, BOOL on)
{
	char s[8];
	BOOL rc = TRUE;

	s[0] = 'B';
	s[1] = '0' + ndx;
	s[2] = on ? '1' : '0';
	itoan(b, s + 3, 2);
	if (TouchPad.RBExists)
		rc = SendLine(RELAY_ADDRESS, s, 5) == 'A';
	return rc;
}

BOOL ToneMode(int ndx, int tone)
{
	char s[8];
	BOOL rc = TRUE;

	s[0] = 'T';
	s[1] = '0' + ndx;
	itoan(tone, s + 2, 2);
	if (TouchPad.RBExists)
		rc = SendLine(RELAY_ADDRESS, s, 4) == 'A';
	return rc;
}

static void ProcessChar(char c)
{
	switch(rxState)
	{
		case 1:
			if (c == MY_ADDRESS)
				rxState++;
			else
				rxState = 0;
			break;
		case 2:
			rxResponse = c;
			rxState++;
			break;
		case 3:
			if (c == '|' || c == '^')
			{
				if (fldPtr < MAX_FIELDS)
				{
					rxBuf[rxPtr++] = '\0';
					fields[fldPtr++] = rxBuf + rxPtr;
				}
			}
			else
			{
				if (c == EOT)
				{
					rxBuf[rxPtr] = '\0';
					rxDone = TRUE;
					rxState = 0;
				}
				else
				{
					if (rxPtr < RX_MAX)
						rxBuf[rxPtr++] = c;
					else
						rxState = 0;		// line too long!
				}
			}
			break;
	}
	if (c == SOH)
	{
		rxPtr = 0;
		fldPtr = 1;
		fields[0] = rxBuf;
		rxState = 1;
	}
}

void __ISR(_UART_1_VECTOR, ipl4) IntUart1Handler(void)
{
	char c;

	if (INTGetFlag(INT_SOURCE_UART_ERROR(UART1)))
	{
		c = U1STA;
		U1STAbits.OERR = 0;
		INTClearFlag(INT_SOURCE_UART_ERROR(UART1));
	}

	if (INTGetFlag(INT_SOURCE_UART_TX(UART1)))
	{
		INTClearFlag(INT_SOURCE_UART_TX(UART1));
	}

	// Is this an RX interrupt?
	if(INTGetFlag(INT_SOURCE_UART_RX(UART1)))
	{
		while (UARTReceivedDataIsAvailable(UART1))
		{
			c = UARTGetDataByte(UART1) & 0x7f;
			ProcessChar(c);
		}
		// Clear the RX interrupt Flag
		INTClearFlag(INT_SOURCE_UART_RX(UART1));
	}
}

