/*
	lamps.c rb June 2,2014
	use SPI to latch up to 32 LED's
	uses TB62706 constant current shift register
*/
#define _SUPPRESS_PLIB_WARNING 1 

#include "GenericTypeDefs.h"
#include "HardwareProfile.h"
#include "lamps.h"

/* globals */
volatile int btBrightness;		// 2 is bright, no PWM
volatile int blBrightness;		// 4 is bright
volatile BOOL backLight;		// true if back light on

/* locals */
static volatile unsigned int currLamp;
static volatile unsigned int lastLamp;
static volatile int blCount;
static volatile int blSetpoint;
static volatile BOOL blToggle;

// back light PWM table
static int blOff[] = { 1, 2, 3, 4, 5 };
static int blOn[] = { 140, 130, 120, 110, 100 };

static volatile int btCount;
static volatile int btSetpoint;
static volatile BOOL btToggle;
static volatile BOOL strobe;

// key light PWM table
static int btOn[] = { 25, 12, 0 };
static int btOff[] = {4, 5, 20 };

#define	ENB	LATDbits.LATD5
#define	LAT	LATDbits.LATD4

static unsigned int keyLamps[] = {
#ifdef RDS16
	0x40,
#endif
#ifdef AST5500
	0x20, 0x100, 0x40,
#endif
#ifndef RDS17
	0x80,
#endif
#ifndef F5TP40ROT
	0x1, 0x2, 0x4, 0x8, 0x10,
#endif
	0x20, 
#ifdef RDS16
	0x1000000,
#elif RDS17
	0x1000000,
#endif
	0x200, 0x400, 0x800, 0x1000,
	0x10000000, 0x8000000, 0x4000000, 0x1000000, 0x2000000, 0x80000000, 0x40000000, 0x20000000,
	 0x40  // this is the extra slide switch lamp when no slide switch
	};

static unsigned int slideLamps[] = {
#ifdef F5TP40WS
	 0x200000, 0x400000, 0x800000   
#elif F5TP40ROT
	// dummy values since no LED's on this version
	 0x200000, 0x200000, 0x200000, 0x200000, 0x200000, 0x200000, 0x200000  
#else
	 0x800000, 0x400000, 0x200000 
#endif
	};

static unsigned int arrowLamps[] = {
	0x40000, 0x80000, 0x100000 
};
// for back light, we don't want the arrow lamps on
#define	ALL_ARROW	0x1c0000
	
static void UpdateLamps(int lamp)
{
	SPI1BUF = lamp;		// change the data, won't latch in till next interrupt
	lastLamp = lamp;
	strobe = TRUE;
}

void LampInit()
{
	TRISDbits.TRISD5 = 0;
	ENB = 0;		// enable current output
	TRISDbits.TRISD4 = 0;
	LAT = 0;		// latch the current data

    SPI1STAT = 0x0000;               // power on state 
	SPI1CON = 0x0000;				// legacy mode
    SPI1CONbits.MODE32 = 1;		// 32 bit mode
	SPI1CONbits.MODE16 = 0;
	SPI1CONbits.MSTEN = 1;			// master mode
	SPI1CONbits.CKP = 0;			// this is the polarity, should be ok?
   	SPI1CONbits.CKE = 1;
	SPI1CONbits.ENHBUF = 1;
	SPI1BRG = 25;					// 1 mhz

	SPI1CONbits.ON = 1;			// enable synchronous serial port
	LampsOff();
	blToggle = FALSE;
	btToggle = FALSE;
	btBrightness = 2;
	blBrightness = 0;
	backLight = FALSE;
	blCount = 0;
	blSetpoint = blOn[blBrightness];
	btCount = 0;
	btSetpoint = btOn[btBrightness];
	currLamp = 0;

	// This statement says: turn on timer3 | have it use an internal clock source | have it
	// use a prescaler of 1:256, and use a period of 0xFFFF or 2^16 cycles
	OpenTimer4( T4_ON | T4_SOURCE_INT | T4_PS_1_256, 30);
	
	UpdateLamps(0);		// turn off all lamps

	// This statement configured the timer to produce an interrupt with a priority of 3
	ConfigIntTimer4( T4_INT_ON | T4_INT_PRIOR_3);
}

/****************************************************************************
  Function:
    void __attribute__((__interrupt__, auto_psv)) _T4Interrupt(void)

  Description:
    Timer ISR, used for PWM for brightness
  Precondition:
    None

  Return Values:
    None

  ***************************************************************************/

void __ISR(_TIMER_4_VECTOR, ipl3) handlesTimer4Ints(void)
{
	unsigned int lmp;
	
	// we're strobing in the last change!
	if (strobe)
	{
		LAT = 1;		// strobe in the previous data
		Nop();
		Nop();
		Nop();
		LAT = 0;		// shift register latch
		strobe = FALSE;
	}

	lmp = lastLamp;
	if (++blCount >= blSetpoint)
	{
		blCount = 0;
		if (blToggle)
		{
			blSetpoint = blOn[blBrightness];
			// this will turn off all the keys that are not currently on
			lmp &= currLamp;		// this will turn off all but the lamps in currLamp that are on
		}
		else
		{
			blSetpoint = blOff[blBrightness];
			if (backLight)
			{
				// only turn on the lamps that are NOT currently on
				lmp = (0xffffffff & ~(currLamp | ALL_ARROW)) | (lmp & currLamp);
			}
		}
		blToggle = ! blToggle;
	}
	if (++btCount >= btSetpoint)
	{
		btCount = 0;
		if (btToggle)
		{
			btSetpoint = btOn[btBrightness];
			lmp &= ~currLamp;
			btToggle = FALSE;
		}
		else
		{
			btSetpoint = btOff[btBrightness];
			// OR in the current lamps to light
			lmp = currLamp | (lmp & ~currLamp);
			if (btOn[btBrightness])
				btToggle = TRUE;
		}
	}
	if (lmp != lastLamp)
		UpdateLamps(lmp);
	 mT4ClearIntFlag();
} // END Timer4 ISR


void LampsOff()
{
	currLamp = 0;
}

void KeyLamp(int lamp, BOOL on)
{
	if (on)
		currLamp |= keyLamps[lamp];
	else
		currLamp &= ~keyLamps[lamp];
}

void DisplayLamp(BOOL on)
{
	if (on)
#ifdef RDS17
		currLamp |= keyLamps[6] | keyLamps[7] | keyLamps[8];
#elif RDS16
		currLamp |= keyLamps[8] | keyLamps[9] | keyLamps[10];
#elif F5TP40ROT
		currLamp |= keyLamps[2] | keyLamps[3] | keyLamps[4];
#elif AST5500
		currLamp |= keyLamps[2] | keyLamps[3] | keyLamps[4];
#else
		currLamp |= keyLamps[7] | keyLamps[8] | keyLamps[9];
#endif
	else
#ifdef RDS17
		currLamp &= ~(keyLamps[6] | keyLamps[7] | keyLamps[8]);
#elif RDS16
		currLamp &= ~(keyLamps[8] | keyLamps[9] | keyLamps[10]);
#elif F5TP40ROT
		currLamp &= ~(keyLamps[2] | keyLamps[3] | keyLamps[4]);
#elif AST5500
		currLamp &= ~(keyLamps[2] | keyLamps[3] | keyLamps[4]);
#else
		currLamp &= ~(keyLamps[7] | keyLamps[8] | keyLamps[9]);
#endif
}

void ErrLamp(int err, BOOL on)
{
	int i, msk;

	if (on)
	{
		for (i = 0, msk = 1; i < 4; i++, msk <<= 1)
		for (i = 0, msk = 1; i < 4; i++, msk <<= 1)
			if (err & msk)
#ifdef F5TP40ROT
				currLamp |= keyLamps[i + 8];
#elif AST5500
				currLamp |= keyLamps[i + 5];
#else
				currLamp |= keyLamps[i];
#endif
	}
	else
#ifdef F5TP40ROT
		currLamp &= ~(keyLamps[8] | keyLamps[9] | keyLamps[10] | keyLamps[11]);
#elif AST5500
		currLamp &= ~(keyLamps[5] | keyLamps[6] | keyLamps[7] | keyLamps[8]);
#else
		currLamp &= ~(keyLamps[0] | keyLamps[1] | keyLamps[2] | keyLamps[3]);
#endif
}

void SlideLamp(int lamp, BOOL on)
{
	currLamp &= ~ (slideLamps[0] | slideLamps[1] | slideLamps[2]);
	if (on && lamp > 0)
		currLamp |= slideLamps[lamp-1];
}

void ArrowLamp(int lamp)
{
	int i, msk;
	for (i = 0, msk = 1; i < 3; i++, msk <<= 1)
		if (lamp & msk)
			currLamp |= arrowLamps[i];
		else
			currLamp &= ~arrowLamps[i];
}
