/*
	logic.c
*/
#define _SUPPRESS_PLIB_WARNING 1 

#include "GenericTypeDefs.h"
#include "HardwareProfile.h"
#include "logic.h"
#include "keys.h"
#include "lamps.h"
#include "flash.h"
#include "comm.h"
#include "display.h"
#include "util.h"

TOUCHPAD TouchPad;
unsigned int mask = 1;
int rawCurrent = 512;		// used for system current
int rawVoltage = 0;		// system voltage
double temperature[ANALOG_CARDS] = { 0, 0, 0 };	// calibrated temperature X 3
int rawTemp[ANALOG_CARDS] = { 0, 0, 0 };		// raw temperature

static unsigned int toneBits[2] = { 0, 0 };		// keep track of what tones are on and what are not
DWORD lastRelayInputs = 0xffffffff;
DWORD lastLightBarInputs = 0xffffffff;
static int voltDelay = 0;
static int temperatureDelay = 0;
static BOOL ignActive = FALSE;			// ignition input is active
static BOOL pbActive = FALSE;			// parking brake input is active
static int ignTimer = 0;
// stops recursion!
#define	MAX_KEY_LIST	64
static int keyList[MAX_KEY_LIST];
static int keyCount;
static int lastStandbyTone;
static BOOL standbyOn = FALSE;

BOOL ValidTouchPad()
{
	return (FlashTouchPad.chkByte1 == 0x58 && FlashTouchPad.chkByte2 == 0x9b);
}

static void ChangeOutputState(INPUT_DATA *inp, BOOL activate)
{
	int ondx, i;

	ondx = 0;
	for(i = 0; i < OUTPUT_COUNT; i++, ondx++)
	{
		if (inp->Outputs[i])
			Relay(ondx, activate);
	}
	// is exp1 installed?
	if (TouchPad.Exp[0])
	{
		for(i = 0; i < EXP_OUTPUTS; i++, ondx++)
		{
			if (inp->OutExp1[i])
				Relay(ondx, activate);
		}
	}
	else
		ondx += EXP_OUTPUTS;
	// is exp2 installed
	if (TouchPad.Exp[1])
	{
		for(i = 0; i < EXP_OUTPUTS; i++, ondx++)
		{
			if (inp->OutExp2[i])
				Relay(ondx, activate);
		}
	}
}

static void SetTone(int ndx, int tone)
{
	if (tone == 6)		// multi-tone?
	{
		ToneMode(ndx, 1);		// tone mode multi
		SoundControl(ndx, 1);	// radio
	}
	else
	{
		ToneMode(ndx, 0);		// tone mode radio
		SoundControl(ndx, tone);
	}
}

/*
	find which light bar program to run if any
*/
static void ChangeLightbar(INPUT_DATA *inp, BOOL activate)
{
	// handle custom text messages
	if (inp->DispFlag)
	{	
		if (activate)
			SetTraffic(CUSTOM, inp->DispMesg);
		else
			ResetTraffic(CUSTOM, inp->DispMesg);
	}
	// valid light bar program?
	if (inp->LightBarPrograms[inp->LightIndex] >= 0 && TouchPad.LBExists)
	{
		if (activate)
		{
			SetTraffic(inp->KeyDisplay, inp->KeyMesg);
			RunLightProgram(LIGHT_ADDRESS, inp->LightBarPrograms[inp->LightIndex], inp->Priority, inp->Dimmed, inp->CanRunFast);
		}
		else
		{
			ResetTraffic(inp->KeyDisplay, inp->KeyMesg);
			StopLightProgram(LIGHT_ADDRESS, inp->LightBarPrograms[inp->LightIndex]);
		}
	}
	// valid traffic director program
	if (inp->TrafficProgram >= 0 && TouchPad.RBExists)
	{
		if (activate)
		{
			SetTraffic(inp->KeyDisplay, inp->KeyMesg);
			RunLightProgram(RELAY_ADDRESS, inp->TrafficProgram, inp->TrafficPriority, FALSE, FALSE);
		}
		else
		{
			ResetTraffic(inp->KeyDisplay, inp->KeyMesg);
			StopLightProgram(RELAY_ADDRESS, inp->TrafficProgram);
		}
	}
	// handle software dimmer
	if (inp->SoftwareDimmer && TouchPad.LBExists)
		SoftwareDimmer(LIGHT_ADDRESS, activate);
	if (inp->FastPatterns && TouchPad.LBExists)
		FastPatterns(LIGHT_ADDRESS, activate);

}

static void OneTone(int ndx, INPUT_DATA *inp, BOOL activate)
{
	int i, msk;

	if ((ndx && inp->SirenTone[ndx] > 0) || (! ndx && inp->SirenMode > 0))
	{
		if (inp->SirenTone[ndx] > 6)		// is it air horn or manual?
		{
			if (ndx == 0 && inp->SirenTone[ndx] >= 9)	// standby mode requested?
			{
				standbyOn = activate;
				lastStandbyTone = inp->SirenTone[ndx] - 6;
			}
			SoundBit(ndx, inp->SirenTone[ndx] - 6, activate);
			// if always in standby, revert back to the correct standby tone
			if ((TouchPad.AlwaysStandby || standbyOn) && ! activate)
			{
				if (standbyOn)
					SoundBit(0, lastStandbyTone, TRUE);
				else
					SoundBit(0, TouchPad.AlwaysStandby, TRUE);
			}
		}
		else
		{
			if (activate)
			{
				SetTone(ndx, inp->SirenTone[ndx]);
				toneBits[ndx] |= 1 << inp->SirenTone[ndx];
			}
			else
			{
				// turn off this tone bit
				toneBits[ndx] &= ~(1 << inp->SirenTone[ndx]);
				// check if any other tones were turned on
				if (toneBits[ndx])
				{
					// figure out what tone to turn on next	
					i = 0;
					for(msk = 1; (msk & toneBits[ndx]) == 0; msk <<= 1)
						i++;
					SetTone(ndx, i);		// this is the next tone to turn on
				}
				else
				{
					SoundControl(ndx, 0);		// turn off tones (other than manual or air horn)
					if (TouchPad.AlwaysStandby || standbyOn)
					{
						if (standbyOn)
							SoundBit(0, lastStandbyTone, TRUE);
						else
							SoundBit(0, TouchPad.AlwaysStandby, TRUE);
					}
				}
			}
		}
	}
}

static void ChangeTone(INPUT_DATA *inp, BOOL activate)
{
	OneTone(0, inp, activate);
	OneTone(1, inp, activate);
}

/*
	new, check for a functional condition
*/
static BOOL ConditionFunc(INPUT_DATA *inp)
{
	BOOL rc = TRUE;
	
	// -1 is ignore
	if (inp->FuncCondition >= 0)
	{
		// even values are active, odd are inactive
		int key = inp->FuncCondition / 2;
		BOOL act = (inp->FuncCondition & 1) == 0;
		if (key < KEY_COUNT)	// double check
		{
			if (act)
			{
				// must be active
				if (! TouchPad.KeyData[key].Active)
					rc = FALSE;
			}
			else
			{
				// must be inactive
				if (TouchPad.KeyData[key].Active)
					rc = FALSE;
			}
		}
	}

	return rc;
}

/*
	check ignition and pb condition
	return true if ok
*/
static BOOL ConditionOk(INPUT_DATA *inp)
{
	BOOL rc = TRUE;
	
	// ignore the condition input if we're deactivating
	// check ignition condition
	if (inp->IgnCondition != IGNORE)
	{
		if ((inp->IgnCondition == INACTIVE && ignActive) || (inp->IgnCondition == ACTIVE && ! ignActive))
			rc = FALSE;
	}
	// check parking brake condition
	if (inp->PBCondition != IGNORE)
	{
		if ((inp->PBCondition == INACTIVE && pbActive) || (inp->PBCondition == ACTIVE && ! pbActive))
			rc = FALSE;
	}
	// added new condition on a function key
	if (! ConditionFunc(inp))
		rc = FALSE;

	return rc;
}

static BOOL ConditionOkActive(INPUT_DATA *inp, BOOL active)
{
	BOOL rc = TRUE;

	if (active)
		rc = ConditionOk(inp);
	return rc;
}

static void SetState(INPUT_DATA *inp, BOOL activate)
{
	// only change the state if there different
	if (activate != inp->Active)
	{
		inp->Active = activate;		// set this key to activated/deactivated
		inp->RunTimer = 0;		// reset the run timer
		ChangeOutputState(inp, activate);
		ChangeTone(inp, activate);
		ChangeLightbar(inp, activate);
	}
}

static void SetStateCond(INPUT_DATA *inp, BOOL activate)
{
	int i;

	// ignore condition on going off!
	if (ConditionOk(inp) || ! activate)
	{
		SetState(inp, activate);
		// recursion
/*		if (key != -1)
			for(i = 0; i < VIRT_COUNT; i++)
				if ((i != key) && (inp->VirtButtons[i] != STATE_NONE))
					SetStateCond(i, &TouchPad.VirtData[i], activate); */
	}
}

static void SetKeyState(int key, INPUT_DATA *inp, BOOL activate)
{
	KeyLamp(key, activate);		// turn on/off this key lamp
	SetState(inp, activate);
}

static void SetKeyStateCond(int key, INPUT_DATA *inp, BOOL activate)
{
	int i;
	// note: ignore condition on going off!
	if (ConditionOk(inp) || ! activate)
	{
		SetKeyState(key, inp, activate);
		// recursion
/*		for(i = 0; i < KEY_COUNT; i++)
			if ((i != key) && (inp->Buttons[i] != STATE_NONE))
				SetKeyStateCond(i, &TouchPad.KeyData[i], activate); */
	}
}

static void AddKey(int key)
{
	if (keyCount < MAX_KEY_LIST)
		keyList[keyCount++] = key;
}

static BOOL InKeyList(int key)
{
	int i;
	BOOL rc;

	for(i = 0; i < keyCount; i++)
		if (keyList[i] == key)
			break;
	
	return i < keyCount;
}


/*
	used only by Inputs to toggle the inputs that are on, to off
	DO NOT CALL THIS! call ButtonLoopOffStart()
*/
static void ButtonLoopOff(INPUT_DATA *i, int ignoreKey)
{
	int key;
	enum OutState *states;

	AddKey(ignoreKey);
	states = i->Buttons;
	for(key = 0; key < KEY_COUNT; key++, states++)
		if (*states == STATE_ON)
		{
			if (! InKeyList(key))
			{
				SetKeyStateCond(key, &TouchPad.KeyData[key], FALSE); 
				ButtonLoopOff(&TouchPad.KeyData[key], key);
			}
		}

	states = i->VirtButtons;
	for(key = 0; key < VIRT_COUNT; key++, states++)
		if (*states == STATE_ON)
		{
			if (! InKeyList(key + 100))
			{
				SetStateCond(&TouchPad.VirtData[key], FALSE); 
				ButtonLoopOff(&TouchPad.VirtData[key], key + 100);
			}
		}
}

/*
	loop through all the key functions including virtual
	DO NOT CALL THIS! call ButtonLoopStart()
*/
static void ButtonLoop(INPUT_DATA *i, int ignoreKey)
{
	int key;
	enum OutState *states;
	BOOL setkey;

	AddKey(ignoreKey);
	states = i->Buttons;
	for(key = 0; key < KEY_COUNT; key++, states++)
		if (*states != STATE_NONE)
		{
			// recursion!
			if (! InKeyList(key))
			{
				SetKeyStateCond(key, &TouchPad.KeyData[key], *states == STATE_ON); 
				ButtonLoop(&TouchPad.KeyData[key], key);
			}
		}

	states = i->VirtButtons;
	for(key = 0; key < VIRT_COUNT; key++, states++)
		if (*states != STATE_NONE)
		{
			// recursion!
			if (! InKeyList(key + 100))
			{
				SetStateCond(&TouchPad.VirtData[key], *states == STATE_ON); 
				ButtonLoop(&TouchPad.VirtData[key], key + 100);
			}
		}
}

static void ButtonLoopStart(INPUT_DATA *i, int ignoreKey)
{
	keyCount = 0;
	ButtonLoop(i, ignoreKey);
}

static void ButtonLoopOffStart(INPUT_DATA *i, int ignoreKey)
{
	keyCount = 0;
	ButtonLoopOff(i, ignoreKey);
}

static void HandleKeyChange(int key, INPUT_DATA *inp, BOOL pressed)
{
	if (! pressed)
	{
		if (inp->KeyType == MOMENTARY)
		{
			if (inp->Active)
				ButtonLoopOffStart(inp, key);
			SetKeyStateCond(key, inp, FALSE);
		}
	}
	else
	{
		if ((inp->KeyType == TOGGLE || inp->KeyType == TIMER) && inp->Active)
		{
			ButtonLoopOffStart(inp, key);
			SetKeyStateCond(key, inp, FALSE);		
		}
		else
		{
			ButtonLoopStart(inp, key);
			SetKeyStateCond(key, inp, TRUE);
		}
	}
}

static void HandleChangeEx(int key, INPUT_DATA *inp, BOOL pressed)
{
	if (! pressed)
	{
		if (inp->KeyType == MOMENTARY)
		{
			if (inp->Active)
				ButtonLoopOffStart(inp, key);
			SetStateCond(inp, FALSE);
		}
	}
	else
	{
		if (inp->KeyType == TOGGLE && inp->Active)
		{
			ButtonLoopOffStart(inp, key);
			SetStateCond(inp, FALSE);		
		}
		else
		{
			ButtonLoopStart(inp, key);
			SetStateCond(inp, TRUE);
		}
	}
}

static void HandleChange(INPUT_DATA *inp, BOOL pressed)
{
	HandleChangeEx(-1, inp, pressed);
}

static int SetKeysOff(INPUT_DATA *i)
{
	int lastKeyOn, key;
	INPUT_DATA *inp;
	enum OutState *states;

	lastKeyOn = -1;

	states = i->Buttons;
	for(key = 0; key < KEY_COUNT; key++, states++)
		if (*states != STATE_NONE)
		{
			inp = &TouchPad.KeyData[key];
			if (inp->Active)
			{
				lastKeyOn = key;
				ButtonLoopOffStart(inp, -1);
				SetKeyStateCond(key, inp, FALSE); 
			}
		}

	states = i->VirtButtons;
	for(key = 0; key < VIRT_COUNT; key++, states++)
		if (*states != STATE_NONE)
		{
			inp = &TouchPad.VirtData[key];
			if (inp->Active)
			{
				lastKeyOn = key + 100;
				ButtonLoopOffStart(inp, -1);
				SetStateCond(inp, FALSE); 
			}
		}
	return lastKeyOn;
}

void AllKeysOff()
{
	INPUT_DATA *inp = TouchPad.KeyData;
	int key;

	for(key = 0; key < KEY_COUNT; key++, inp++)
		if (inp->Active)		// is this timer active?
		{
			SetKeyStateCond(key, inp, FALSE);
			SetKeysOff(inp);
		}
}


static BOOL SetNextKeyOn(INPUT_DATA *i, int lastKeyOn)
{
	BOOL nextKey = FALSE;
	BOOL setKey = FALSE;
	int key;
	INPUT_DATA *inp;
	enum OutState *states;
	
	if (lastKeyOn >= 100)		// starts at virtual?
	{
		states = i->VirtButtons; 
		for(key = 0; key < VIRT_COUNT; key++, states++)
			if (*states != STATE_NONE)
			{
				if (nextKey)		// found a key to turn on
				{
					inp = &TouchPad.VirtData[key];
					HandleChangeEx(key + 100, inp, TRUE);
//					SetStateCond(inp, TRUE);
					setKey = TRUE;
					break;
				}
				else
					if (key + 100 == lastKeyOn)
						nextKey = TRUE;
			}
	}
	else
	{
		states = i->Buttons; 
		for(key = 0; key < KEY_COUNT; key++, states++)
			if (*states != STATE_NONE)
			{
				if (nextKey)		// found a key to turn on
				{
					inp = &TouchPad.KeyData[key];
					HandleKeyChange(key, inp, TRUE);
//					SetKeyStateCond(key, inp, TRUE);
					setKey = TRUE;
					break;
				}
				else
					if (key == lastKeyOn)
						nextKey = TRUE;
			}
		// didn't find one on the non-virt keyboard but the next one is the virtual
		if (! setKey && nextKey)
		{
			states = i->VirtButtons; 
			for(key = 0; key < VIRT_COUNT; key++, states++)
				if (*states != STATE_NONE)
				{
					inp = &TouchPad.VirtData[key];
					HandleChangeEx(key + 100, inp, TRUE);
//					SetStateCond(inp, TRUE);
					setKey = TRUE;
					break;
				}
		}
	}
	return setKey;
}

static BOOL SetFirstKey(INPUT_DATA *i)
{
	INPUT_DATA *inp;
	int key;
	enum OutState *states;
	BOOL rc = FALSE;

	states = i->Buttons;
	for(key = 0; key < KEY_COUNT; key++, states++)
		if (*states != STATE_NONE)		// found the 1st key to turn on
		{
			inp = &TouchPad.KeyData[key];
			HandleKeyChange(key, inp, TRUE);
//			SetKeyStateCond(key, inp, TRUE);
			rc = TRUE;
			break;
		}
	if (! rc)		// didn't find in the non-virt keys, check virtual
	{
		states = i->VirtButtons;
		for(key = 0; key < VIRT_COUNT; key++, states++)
			if (*states != STATE_NONE)		// found the 1st key to turn on
			{
				inp = &TouchPad.VirtData[key];
				HandleChangeEx(key + 100, inp, TRUE);
//				SetStateCond(inp, TRUE);
				rc = TRUE;
				break;
			}
	}

	return rc;
}

/*
	check if cycle type key
*/
static BOOL CheckForCycle(int key, INPUT_DATA *inp, BOOL down)
{
	BOOL rc = FALSE;
	int lastKey;

	if (inp->KeyType == CYCLE || inp->KeyType == CYCLEOFF)
	{
		rc = TRUE;
		if (down)
		{
			// check if the cycle is active or not
			if (inp->Active || inp->KeyType == CYCLE)
			{
				lastKey = SetKeysOff(inp);
				if (lastKey == -1)		// no button was on?
				{
					SetFirstKey(inp);
				}
				else
				{
					if (! SetNextKeyOn(inp, lastKey))		// try and set the next key
					{
						if (inp->KeyType == CYCLE)
							SetFirstKey(inp);
						else
						{
							ButtonLoopOffStart(inp, -1);
							if (key == -1)
								SetStateCond(inp, FALSE);		// hardware input
							else
								SetKeyStateCond(key, inp, FALSE);		// cycle off, so turn off
						}
					}
				}
			}
			else
			{
				// cycle off type
				if (key == -1)
					HandleChange(inp, TRUE);		// hardware input
				else
					HandleKeyChange(key, inp, TRUE);
				SetKeysOff(inp);
				SetFirstKey(inp);
			}
		}
	}
	return rc;
}

void KeyPressEvent(int key, BOOL down)
{
	INPUT_DATA *inp;

	// do we have flash data for touch pad?
	if (! ValidTouchPad())
		return;
	inp = &TouchPad.KeyData[key];
	if (! CheckForCycle(key, inp, down))
	{
		HandleKeyChange(key, inp, down);
	}
}

/*
	this event will occur when a key is held for two seconds
*/
void KeyHeldEvent(int key)
{
	INPUT_DATA *inp = &TouchPad.KeyData[key];

	if (inp->KeyType == CYCLEOFF)		// check if it's a cycle off key
	{
		SetKeysOff(inp);
		SetKeyStateCond(key, inp, FALSE);		// key was held for 2 secs, so turn off
	}
}

void SlideEvent(int pos, BOOL on)
{
	// do we have flash data for touch pad?
	if (! ValidTouchPad())
		return;

	SlideLamp(pos, on);
	if (on)
		ButtonLoopStart(&TouchPad.SlideData[pos], -1);
	HandleChange(&TouchPad.SlideData[pos], on);
}

static BOOL IsInputActive(int key, INPUT_DATA *inp, int typ)
{
	DWORD input;
	BOOL rc = FALSE;
	BOOL toggle = FALSE;

	switch(typ)
	{
		case 0:
			rc = IsSlideOn(key);	// is slide input physically on?
			break;
		case 1:
		case 2:
			if (typ == 1)
			{
				// special for ign/pk inactive keys
				if (key >= 9)
				{
					key -= 9;
					toggle = TRUE;
				}
				input = (lastRelayInputs >>= (2 * key)) & 3;
			}
			else
				input = (lastLightBarInputs >>= (2 * key)) & 3;
			if (input == 1 && inp->InputTrigger == NEGATIVE)
				rc = TRUE;
			else
				if (input == 2 && inp->InputTrigger == POSITIVE)
					rc = TRUE;
			break;
		case -1:
			rc = FALSE;		// for virtual buttons
			break;
	}
	// for inactive IGN/PK
	if (toggle)
		rc = ! rc;
	return rc;
}

/*
	deactivate anything that is active if the ignition input or pb inputs deactivate
	and those functions are checking the ignition input condition
*/

static void HandleConditionChangeX(INPUT_DATA *inp, int cnt, BOOL isKey, int typ)
{
	int key;

	for(key = 0; key < cnt; key++, inp++)
		if (inp->IgnCondition != IGNORE || inp->PBCondition != IGNORE)		// is it conditional on ign or pb?
		{
			if (ConditionOk(inp))		// has the ign or pb condition gone away?
			{
				if (! inp->Active)	// if this input is not active
				{
					// we only want to turn on inputs and slide, no keys!
					if (inp->KeyType == MOMENTARY)		// monetary types only!
						if (IsInputActive(key, inp, typ))
							HandleChange(inp, TRUE);	// turn it back on!
				}
			}
			else
			{
				if (inp->Active)	// only if active
				{
					// everything goes off if ignition or pb changed
					ButtonLoopOffStart(inp, isKey ? key : -1);
					if (isKey)
						SetKeyState(key, inp, FALSE);
					else
						SetState(inp, FALSE);
				}
			}
		}
}

static void HandleConditionChange()
{
	HandleConditionChangeX(TouchPad.KeyData, KEY_COUNT, TRUE, -1);
	HandleConditionChangeX(TouchPad.SlideData, SLIDE_COUNT, FALSE, 0);
	HandleConditionChangeX(TouchPad.RelayBoxInputs, RELAYBOX_INPUTS, FALSE, 1);
	HandleConditionChangeX(TouchPad.LightBarInputs, TouchPad.LightBarType > 0 ? 4 : LIGHTBAR_INPUTS, FALSE, 2);
	HandleConditionChangeX(TouchPad.VirtData, VIRT_COUNT, FALSE, -1);
}

static void SetIgnPb(int key, BOOL lbFlag, BOOL active)
{
	BOOL changeFlag;

	if (! lbFlag || ! TouchPad.RBExists)		// is relaybox or lightbar
	{
		if (key == 0)		// ignition
		{
			changeFlag = ignActive != active;
			ignActive = active;
			// for now, back light is on when ignition is on
			backLight = active;
			if (changeFlag)
				HandleConditionChange();		// ignition change
		}
		else		// park kill
		{
			changeFlag = pbActive != active;
			pbActive = active;
			if (changeFlag)
				HandleConditionChange();		// pb change
		}
	}
}

/*
	toggle one set of button light bar patterns
*/
static ToggleOne(INPUT_DATA *inp, int cnt, BOOL *SaveIt)
{
	int key, currNdx;

	for(key = 0; key < cnt; key++, inp++)
		if (inp->Active)	// if this input is active
			if (inp->LightBarPrograms[inp->LightIndex] >= 0)	// is there a light bar program running?
			{
				// stop the current pattern	
				currNdx = inp->LightIndex;			
				StopLightProgram(LIGHT_ADDRESS, inp->LightBarPrograms[inp->LightIndex]);
				// find the next pattern
				while(++inp->LightIndex < MAX_PROGS)
					if (inp->LightBarPrograms[inp->LightIndex] >= 0)	// did we find one?
						break;
				if (inp->LightIndex >= MAX_PROGS)	// no, reset back to zero
					inp->LightIndex = 0;
				// save this pattern selection
				if (inp->LightIndex != currNdx)
					*SaveIt = TRUE;
				// start the next pattern
				RunLightProgram(LIGHT_ADDRESS, inp->LightBarPrograms[inp->LightIndex], inp->Priority, inp->Dimmed, inp->CanRunFast);
			}
}

/*
	toggle input is pressed, cycle to next pattern
*/
static void TogglePattern()
{
	BOOL SaveIt = FALSE;

	ToggleOne(TouchPad.KeyData, KEY_COUNT, &SaveIt);
	ToggleOne(TouchPad.RelayBoxInputs, RELAYBOX_INPUTS, &SaveIt);
	ToggleOne(TouchPad.LightBarInputs, TouchPad.LightBarType > 0 ? 4 : LIGHTBAR_INPUTS, &SaveIt);
	ToggleOne(TouchPad.VirtData, VIRT_COUNT, &SaveIt);
	ToggleOne(TouchPad.SlideData, SLIDE_COUNT, &SaveIt);
	if (SaveIt)
		SaveTouchPad(&TouchPad, sizeof(TouchPad));
}

static void HandleTemperature(int key, INPUT_DATA *inp)
{
	BOOL active;

	active = FALSE;
	if (inp->TempActiveAbove)
	{
		if (temperature[key] > inp->TempSetpoint)
			active = TRUE;
	}
	else
	{
		if (temperature[key] < inp->TempSetpoint)
			active = TRUE;
	}
	if (active)
	{
		if (! inp->Active)
		{
			SetStateCond(inp, TRUE);
			ButtonLoopStart(inp, -1);
		}
	}
	else
	{
		if (inp->Active)
		{
			SetStateCond(inp, FALSE);
			ButtonLoopOffStart(inp, -1);
		}
	}
}

static void HandleInputs(DWORD input, DWORD lastInput, INPUT_DATA *inp, int num, BOOL lbFlag)
{
	int key;
	DWORD saveInput, saveLast;
	INPUT_DATA *inps, *inpt;

	inps = inp;
	saveInput = input;
	saveLast = lastInput;

	for(key = 0; key < num; key++, lastInput >>= 2, input >>= 2, inp++)
		if ((input & 3) != (lastInput & 3))		// change of state?
		{
			switch(input & 3)
			{
				case 0:			// open or disconnected
					if (! lbFlag || key != num - 1)		// ignore flash pattern toggle
					{
						if (! CheckForCycle(-1, inp, FALSE))
						{
							SetStateCond(inp, FALSE);
							// if this input type is momentary
							// then toggle off any keys that are on
							if (inp->KeyType == MOMENTARY)
								ButtonLoopOffStart(inp, -1);
							// for ignition and pk we have an inactive state we need to handle
							if (! lbFlag && key < 2)
							{
								inpt = &inps[9 + key];
								SetStateCond(inpt, TRUE);
								if (ConditionOk(inpt))
									ButtonLoopStart(inpt, -1);
							}
						}
					}
					break;
				case 1:			// grounded
					if (! lbFlag || key != num - 1)		// ignore flash pattern toggle
					{
						if (! CheckForCycle(-1, inp, inp->InputTrigger == NEGATIVE))
						{
							SetStateCond(inp, inp->InputTrigger == NEGATIVE);
							if (inp->InputTrigger == NEGATIVE && ConditionOk(inp))
								ButtonLoopStart(inp, -1);
							// for ignition and pk we have an inactive state we need to handle
							if (! lbFlag && key < 2)
							{
								inpt = &inps[9 + key];
								SetStateCond(inpt, inp->InputTrigger == POSITIVE);
								if (inp->InputTrigger == POSITIVE && ConditionOk(inpt))
									ButtonLoopStart(inpt, -1);
							}
						}
					}
					break;
				case 2:			// +12v
					if (lbFlag && key == num - 1)		// should be num - 1, input 16 is flash pattern toggle!, it's 4 for the Odyssey bars
						TogglePattern();
					else
					{
						if (! CheckForCycle(-1, inp, inp->InputTrigger == POSITIVE))
						{
							SetStateCond(inp, inp->InputTrigger == POSITIVE);
							if (inp->InputTrigger == POSITIVE && ConditionOk(inp))
								ButtonLoopStart(inp, -1);
							// for ignition and pk we have an inactive state we need to handle
							if (! lbFlag && key < 2)
							{
								inpt = &inps[9 + key];
								SetStateCond(inpt, inp->InputTrigger == NEGATIVE);
								if (inp->InputTrigger == NEGATIVE && ConditionOk(inpt))
									ButtonLoopStart(inpt, -1);
							}
						}
					}
					break;
			}
		}

	// do the ignition and pb logic change of state
	// added support for relay box inactive function keys
	for(key = 0; key < 2; key++, saveLast >>= 2, saveInput >>= 2)
		if ((saveInput & 3) != (saveLast & 3))		// change of state?
			switch(saveInput & 3)
			{
				case 0:
					SetIgnPb(key, lbFlag, FALSE);
					break;
				case 1:
					SetIgnPb(key, lbFlag, inps->InputTrigger == NEGATIVE);
					break;
				case 2:
					SetIgnPb(key, lbFlag, inps->InputTrigger == POSITIVE);
					break;
			}
}

static void HandleTimers(INPUT_DATA *inp, int cnt, BOOL isKey)
{
	int key;

	for(key = 0; key < cnt; key++, inp++)
		if (inp->KeyType == TIMER)		// timer type?
		{
			if (inp->Active)		// is this timer active?
			{
				if (inp->RunTimer >= inp->TimerValue)		// timed out?
				{
					ButtonLoopOffStart(inp, isKey ? key : -1);
					if (isKey)
						SetKeyStateCond(key, inp, FALSE);
					else
						SetStateCond(inp, FALSE);
				}
				else
					inp->RunTimer++;				// add another 100ms to the run timer
			}
		}
}

static void HandleVirtTemperature()
{
	int key;
	INPUT_DATA *inp = TouchPad.VirtData;

	for(key = 0; key < VIRT_COUNT; key++, inp++)
		if (inp->KeyType == TEMP_TRIGGER)		// temperature trigger type?
		{
			HandleTemperature(inp->TempDevice, inp);
		}
}

static void HandleIgnTimeout()
{
	// only if we have a relay box
	if (TouchPad.RBExists)
	{
		// did relay box just power up?
		if (TouchPad.AlwaysStandby)
			if (RelayBoxPowerUp())
			{
				// need to put in standby
				SoundBit(0, TouchPad.AlwaysStandby, TRUE);
			}
		if (! ignActive)
		{
			if (TouchPad.IgnTimeout)
				if (ignTimer++ == TouchPad.IgnTimeout * 600)
					SendPowerOff();
		}
		else
			ignTimer = 0;
	}
}

/*
	called by keytasks() every 100 ms
*/
void TimerEvent()
{
	DWORD inputs;
	int i;
	WORD t;
	double d;

	// do we have flash data for touch pad?
	if (! ValidTouchPad())
		return;

	HandleIgnTimeout();
	HandleTimers(TouchPad.KeyData, KEY_COUNT, TRUE);
	HandleTimers(TouchPad.RelayBoxInputs, RELAYBOX_INPUTS, FALSE);
	HandleTimers(TouchPad.LightBarInputs, TouchPad.LightBarType > 0 ? 4 : LIGHTBAR_INPUTS, FALSE);
	HandleTimers(TouchPad.VirtData, VIRT_COUNT, FALSE);

	if (++temperatureDelay >= 10)		// once per second
	{
		for (i = 0; i < ANALOG_CARDS; i++)
			if (TouchPad.Analog[i])
			{		
				if (ReadTemperature(TEMP_ADDRESS + i))
				{
					t = HexToWord(rxBuf);
					rawTemp[i] = t;		// raw readings
					// 3Deg C = 110 counts
					// 23Deg C = 148 counts = 1.9 counts / deg C
					// Zero = 104.3 
					// therefore temp = counts - 104.3 * (1 / 1.9)
					// temp = counts - 104.3 * .526315
//					d = ((double)t - 104.3) * 0.526315;
					d = ((double)t) * Params.TempCal[i];
					d -= 55.0;		// zero counts is -55C
					// if fahrenheit
					if (! TouchPad.TempC)
						d = d * 1.8 + 32.0;
					temperature[i] = d;
					HandleTemperature(i, &TouchPad.AnalogData[i]);
				}
			}
		temperatureDelay = 0;
	}

	// new, can program virtuals as temperature triggered inputs
	HandleVirtTemperature();

	if (TouchPad.RBExists)
	{
		if (++voltDelay >= 5)
		{
			// for system current calc and voltage
			if (ReadVoltage(RELAY_ADDRESS))
			{
				rawCurrent = atoi(rxBuf);
				rawVoltage = atoi(rxBuf + 5);
			}
			voltDelay = 0;
		}
		if (ReadInputs(RELAY_ADDRESS, 'K'))		// read analog relay box inputs
		{
			inputs = HexToWord(rxBuf) << 16;		// bits 0-15 are inputs 1-8, bit 16,17 is horn ring on, bit 18,19 is horn pulse
			inputs += HexToWord(rxBuf + 4);
			// Note: -2 because last two inputs are ignition condition off and pkg off
			HandleInputs(inputs, lastRelayInputs, TouchPad.RelayBoxInputs, RELAYBOX_INPUTS - 2, FALSE);
			lastRelayInputs = inputs; 
		}
	}
	if (TouchPad.LBExists)
	{
		if (ReadInputs(LIGHT_ADDRESS, 'K'))	// read analog inputs
		{
			inputs = HexToWord(rxBuf) << 16;		// 32 bits for 16 inputs from light bar
			inputs += HexToWord(rxBuf + 4);
			HandleInputs(inputs, lastLightBarInputs, TouchPad.LightBarInputs, TouchPad.LightBarType > 0 ? 4 : LIGHTBAR_INPUTS, TRUE);
			lastLightBarInputs = inputs;
		}
	}
}

void LogicInit()
{
	int i;

	for(i = 0; i < KEY_COUNT; i++)
		key_funcs[i] = KeyPressEvent;
	for(i = 0; i < KEY_COUNT; i++)
		key_held[i] = KeyHeldEvent;
	slide_func = SlideEvent;
	tmr_func = TimerEvent;
	
	// make sure flash memory is valid
	if (ValidTouchPad())
	{
		UserFlashPageReadBytes(&FlashTouchPad, (BYTE *)&TouchPad, sizeof(TouchPad));
		// setup the brightness
		btBrightness = TouchPad.ButtonIntensity;
		blBrightness = TouchPad.BacklightIntensity;
		dispBrightness = TouchPad.DisplayIntensity;
	}
}
