/****************************************************************************
	timer.c  DM&T Services rb May 8,2009
	This unit provides simple timer tick functions with a 30ms time base
	should be more than adequate for our purposes
  ***************************************************************************/
#define _SUPPRESS_PLIB_WARNING 1 

#include "GenericTypeDefs.h"
#include "HardwareProfile.h"
#include "timer.h"
#include "beepalert.h"
	
// *****************************************************************************
// *****************************************************************************
// Constants
// *****************************************************************************
// *****************************************************************************

// We are taking Timer 3 for tick purposes

// NOTE - The datasheet doesn't state this, but the timer does get reset to 0
// after a period register match.  So we don't have to worry about resetting
// the timer manually.

#define	INTERVAL_MS					10

// *****************************************************************************
// *****************************************************************************
// Data
// *****************************************************************************
// *****************************************************************************

static volatile int _ticks;		// use helper functions to be interrupt safe



/****************************************************************************
  Function:
     void InitializeTimer( void )

  Description:
    This function initializes the tick timer.  It configures and starts the
    timer, and enables the timer interrupt.  The interrupt rate will be 30ms.

  Parameters:
    None

  Returns:
    None

  ***************************************************************************/
void Timer3Init( void )
{
	_ticks = 0L;			// reset tick
	_buzzer = 0;

	// This statement says: turn on timer3 | have it use an internal clock source | have it
	// use a prescaler of 1:256, and use a period of 1171
	OpenTimer3( T3_ON | T3_SOURCE_INT | T3_PS_1_256, 1171);
	
	// Timer2 as configured would trigger an interrupt at a frequency of (30MHZ/256/1171), or 100.074
	// times a second.
	
	// This statement configured the timer to produce an interrupt with a priority of 2
	ConfigIntTimer3( T3_INT_ON | T3_INT_PRIOR_2);
}


/****************************************************************************
  Function:
    void __attribute__((__interrupt__, auto_psv)) _T3Interrupt(void)

  Description:
    Timer ISR, used for internal tick purposes.
  Precondition:
    None

  Return Values:
    None

  ***************************************************************************/

void __ISR(_TIMER_3_VECTOR, ipl2) handlesTimer3Ints(void)
{
	// **make sure iplx matches the timers interrupt priority level
	_ticks++;
	// buzzer off logic
	if(_buzzer)
	{
		if (! --_buzzer)
			BUZZER = 0;
	}
	
	mT3ClearIntFlag();
} // END Timer2 ISR

/****************************************************************************
  Function:
     int GetTickCount( void )

  Description:
    This function gets the current tick count and is interrupt safe.
	Don't reference _ticks directly, use this helper function.

  Parameters:
    None

  Returns:
    int value of ticks (30ms time base, zero'd on startup)

  ***************************************************************************/

int GetTickCount( void )
{
    int result;

	IEC0bits.T3IE = 0;		// disable timer interupt 3
	result = _ticks;
	IEC0bits.T3IE = 1;		// enable timer interupt 3

	Nop();				// just in case this is getting called in a tight loop
	Nop();
	return result;		// return with 10ms tick count
}

/****************************************************************************
  Function:
     BOOL HasElapsed(int tick, WORD ms)

  Description:
    This function returns true if ms has elapsed
	Only accurate to a 30ms time base for now

  Parameters:
    tick - value set by GetTickCount() before calling this function
	ms - value in ms period to be checked if elapsed

  Returns:
    true if ms

  ***************************************************************************/

BOOL HasElapsed(int tick, int ms)
{
	int diff = GetTickCount() - tick;
	return diff >= (ms / INTERVAL_MS);
}


/**********************************
  Function
     unsigned int TickDiff(long tick)
  Description:
     pass the tick count from GetTickCount
     returns the 30ms time periods as an unsigned int
*/
unsigned int TickDiff(int tick)
{
	unsigned int Result = GetTickCount() - tick;
	return Result;
}

/****************************************************************************
  Function:
     BOOL Repeat(long *tick, WORD ms)

  Description:
    This function returns true if ms has elapsed
	Only accurate to a 30ms time base for now
	Used for repetive same time tasks

  Parameters:
    tick - should be set to zero initially
	ms - value in ms period to be checked if elapsed

  Returns:
    true if > ms, can be called successively for repetitive tasks

  ***************************************************************************/

BOOL Repeat(int *tick, int ms)
{
	BOOL result;
	long t;

	t = GetTickCount();
	if(! *tick)
		*tick = t;
	result = (int)(t - *tick) >= (ms / INTERVAL_MS);
	if(result)
		*tick = t;		// reset for next time

	return result;
}
